%
% Sample script to convert NIFTI data for tensorreg, 
% adding noise. This script exports the whole 3D volume
% converted to tensors.
%


%
% Set up parameters for script
%

rho=0.001;
iters=5000;
noises=[10 50];
a=0.15;

% Export whole brawin
slice=[];
BASEPREFIX='brain_lownoise';

% Export slice 23 only
%slice=23;
%BASEPREFIX='../brain_lownoise_23';

%
% Load data (NIFTI format)
%

d0=dti_load_nii('../../../dti_lownoise/', ...
                'data_eddycorrect', ...
                '20100121_164850ep2ddiff2b12dgrpa2neus009a001', ...
                slice);

[sig0, f0, mask]=dti_data_to_tensor(d0, 0);

%
% Calculate regularisation parameters.
% (Only a few of these are used at this moment, others
%  are historical and kept here for the time being.)

L=sym2eigs_multi(f0);
sc=max(max(L));

a1=a*sc;
b1=10*a1;
a1p=a*abs(log(sc));

%a2=0.4*a1;
%b2=0.6*a1;
%a2p=0.4*a*abs(log(sc));

%sc_signal=max(flattenT(mean(d.I, 3)));
%a1s=a*sc_signal;
%a2s=0.4*a*sc_signal;


%
% Export noise-free data as raw doubles
%
tensorreg_export([BASEPREFIX, '-f0'], sig0, f0, mask, a1, b1, rho, iters);

%
% Generate and export noisy data as raw doubles
%
for j=1:length(noises)
    noise=noises(j);
    d=d0; 
    d.I=ricernd(d0.I, noise);
    [sig, f]=dti_data_to_tensor(d, 0);
    name=sprintf('%s-f%d', BASEPREFIX, noise);
    tensorreg_export(name, sig0, f, mask, a1, b1, rho, iters);
end

