%
% Generate a synthetic test data of Sym^2(\R^2) tensors as 
% explained in the report "Total generalised variation in 
% diffusion tensor imaging".
%

function [sig, u]=testdata4()
    w=128;
    sig=[2 2 w w];
    l=w*w;
    
    u0=sympres(eye(2));
    ux=sympres([1.1 0; 0 0.9]);
    u=repmat(ux, 1, l);
    
    xs=1:w;
    ys=1:w;
    dx=sympres(zeros(2));
    dy=sympres(zeros(2));
    u=fill(u, w, xs, ys, u0, dx, dy);
    
    xs=[1:64];
    ys=[1:64];
    lambda=0.005;
    dx=sympres(lambda*[0 1; 1 2]);
    dy=sympres(lambda*[-2 -1; -1 0]);
    u=fill(u, w, xs, ys, u0, dx, dy);

    xs=[1:64];
    ys=[65:128];
    lambda=0.02;
    dx=sympres(lambda*[0 0; 0 1]);
    dy=sympres(lambda*[1 0; 0 0]);
    u=fill(u, w, xs, ys, u0, dx, dy);
    
    xs=[65:128];
    ys=[1:64];
    v1=[1;0]/sqrt(2); v2=[0;1]/sqrt(2);
    %u1=sympres(3*v2*v2'+1*v1*v1');
    u1=sympres(1.5*v2*v2'+1*v1*v1');
    u=fill_interpr_x(u, w, xs, ys, u1);
    %u=2*u/max(abs(sym2eigs_multi(u)));
end

function u=fill(u, w, xs, ys, u0, dx, dy)
    z=u0;
    for x=xs
        for y=ys
            u(:, (y-1)*w+x)=z+(y-ys(1))*dy;
        end
        z=z+dx;
    end
        
end

function u=fill_interpr_x(u, w, xs, ys, u1)
    M=symprestomat(u1);
    for x=xs
        for y=ys
            u(:, (y-1)*w+x)=sympres(interpr(M, 0, pi/4, x, xs));
        end
    end
end

function MR=interpr(M, minangle, maxangle, y, ys)
    angle=minangle+(maxangle-minangle)*(y-ys(1))/(ys(end)-ys(1));
    R=[1 0; 0 1]*cos(angle)+[0 -1; 1 0]*sin(angle);
    MR=R*M*R';
end
