%
% dticode/matlab/calc_DIVnonsym.m
%
% Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
%
function M=calc_DIVnonsym(sig)
    order=sig(2);
    if order==1
        M=calc_DIV1nonsym(sig);
    elseif order==2
        M=calc_DIV2nonsym(sig);
    elseif order==3
        M=calc_DIV3nonsym(sig);
    elseif order==4
        M=calc_DIV4nonsym(sig);
    else
        error('calc_DIVnonsym not implemented for given order')
    end
end

% ---

function M=calc_DIV1nonsym(sig)
    dim=sig(1);
    dsig=orderdown(sig);
    ds=tsize_nonsym(dsig);
    ss=tsize_nonsym(sig);
    % x/y/z/etc. differentials in source space after each other
    M=zeros(ds, dim*ss);
    is=0:(dim-1);
    for i1=is
        j=index_ns(dsig, [0]);
        i=index_ns(sig, [i1]);
        k=i+ss*i1; % x/y/z/etc. offset
        M(j, k)=M(j, k)+1;
    end
end

% ---

function M=calc_DIV2nonsym(sig)
    dim=sig(1);
    dsig=orderdown(sig);
    ds=tsize_nonsym(dsig);
    ss=tsize_nonsym(sig);
    % x/y/z/etc. differentials in source space after each other
    M=zeros(ds, dim*ss);
    is=0:(dim-1);
    for i1=is
        for i2=is
            j=index_ns(dsig, [i2]);
            i=index_ns(sig, [i1 i2]);
            k=i+ss*i1; % x/y/z/etc. offset
            M(j, k)=M(j, k)+1;
        end
    end
end

% ---

function M=calc_DIV3nonsym(sig)
    dim=sig(1);
    dsig=orderdown(sig);
    ds=tsize_nonsym(dsig);
    ss=tsize_nonsym(sig);
    % x/y/z/etc. differentials in source space after each other
    M=zeros(ds, dim*ss);
    is=0:(dim-1);
    for i1=is
        for i2=is
            for i3=is
                j=index_ns(dsig, [i2 i3]);
                i=index_ns(sig, [i1 i2 i3]);
                k=i+ss*i1; % x/y/z/etc. offset
                M(j, k)=M(j, k)+1;
            end
        end
    end
end

% ---

function M=calc_DIV4nonsym(sig)
    dim=sig(1);
    dsig=orderdown(sig);
    ds=tsize_nonsym(dsig);
    ss=tsize_nonsym(sig);
    % x/y/z/etc. differentials in source space after each other
    M=zeros(ds, dim*ss);
    is=0:(dim-1);
    for i1=is
        for i2=is
            for i3=is
                for i4=is
                    j=index_ns(dsig, [i2 i3 i4]);
                    i=index_ns(sig, [i1 i2 i3 i4]);
                    k=i+ss*i1; % x/y/z/etc. offset
                    M(j, k)=M(j, k)+1;
                end
            end
        end
    end
end
