%
% dticode/matlab/dti_data_to_tensor.m
%
% Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
%
function [sig, f, mask]=dti_data_to_tensor(d, flip)
    if nargin<2
        flip=0;
    elseif flip~=0
        warning('Flip parameter should normally be zero!');
    end

    w=size(d.I, 1);
    h=size(d.I, 2);
    q=size(d.I, 3);
    m=size(d.I, 4);
    
    sig=[3, 2, w, h, q];
    ts=tsize_sym(sig); % = 6

    ix=(d.I<=0);

    if any(ix>0)
        warning('Negative echo magnitudes! Taking absolute value.')
        d.I=abs(d.I);
    end
    
    B=[];
    bnz=[];
    bz=[];
    bz_current=[];
    %jj=[index_sym(2, [0 0]) index_sym(2, [0 1]) index_sym(2, [1 1])];
    jj=[kx([0 0]) kx([0 1]) kx([1 1]) kx([0 2]) kx([1 2]) kx([2 2])];
    for i=1:m
        if d.bval(i)==0
            bz_current=i;
        else
            assert(~isempty(bz_current));
            bv=d.bvec(:, i);
            bm=d.bval(i)*bv*bv';
            if flip
                %warning('flip now in 3d?')
                bs(jj)=[bm(3, 3) 2*bm(1,3) bm(1,1) 2*bm(2,3) 2*bm(1,2) bm(2,2)];
            else
                bs(jj)=[bm(1, 1) 2*bm(1,2) bm(2,2) 2*bm(1,3) 2*bm(2,3) bm(3,3)];
            end
            B=[B; bs];
            bz=[bz, bz_current];
            bnz=[bnz, i];
        end
    end
    
    u=zeros(ts, w, h);

    f=zeros(ts, w*h*q);
    im=reshape(shiftdim(d.I, 3), m, w*h*q);
    c=log(im(bnz, :)./im(bz, :));
    
    if any(~isreal(c))
        warning('Complex logarithm!');
    end
    
    if any(any(isnan(c)))
        warning('NaN')
    end
    
    ok=all(isreal(c) & isfinite(c), 1);
    f(:, ok)=(-B) \ c(:, ok);
    
    % calculate mask
    avg=reshape(sum(abs(d.I), 4), 1, w*h*q);
    threshold=0.10*max(avg);
    mask=avg>threshold;
end

function res=kx(a)
    res=index_sym(3, a);
end
