%
% dticode/matlab/tensorreg_export_raw.m
%
% Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
%
function tensorreg_export_raw(name, data, initf, scale)
    w=size(data.I, 1);
    h=size(data.I, 2);
    d=size(data.I, 3);
    m=size(data.I, 4);

    infilename=[name, '.doubles'];
    s0filename=[name, '-s0.doubles'];
    initfilename=[name, '-init.doubles'];
    amatname=[name, '-amat.doubles'];
    maskname=[name, '-mask.mat'];
    
    im=reshape(shiftdim(data.I, 3), m, w*h*d);
    
    % Construct A=[B_1; ...; B_n], where
    % B_i=bval(i) * (bvec(i) (*) bvec(i))
    A=[];
    W=[];
    bnz=[];
    bz=[];
    bz_current=[];
    scale_current=1;
    jj=[kx([0 0]) kx([0 1]) kx([1 1]) kx([0 2]) kx([1 2]) kx([2 2])];
    nn=1;
    for i=1:m
        if data.bval(i)==0
            bz_current=i;
            if scale==2
                scale_current=norm(im(i, :), 2);
            end                
        else
            assert(~isempty(bz_current));
            bv=data.bvec(:, i);
            bm=data.bval(i)*bv*bv';
            bs(jj)=[bm(1, 1) 2*bm(1,2) bm(2,2) 2*bm(1,3) 2*bm(2,3) bm(3,3)];
            % Normalise for numerical accuracy.
            if scale==1
                nn=data.bval(i);%norm(bs);
            end
            A=[A; bs/nn];
            W=[W; nn];
            %A=[A; bs];
            bz=[bz, bz_current];
            bnz=[bnz, i];
        end
        im(i, :)=im(i, :)/scale_current;
    end

    bad=~isreal(im) | ~isfinite(im);
    if any(any(bad))
        warning on;
        warning('Complex, NaN or Inf entries -- zeroing');
        im(bad)=0;
    end

    W
    
    % Scaling for numerical accuracy of A
    if scale==1
        wmat=repmat(W, 1, w*h*d);
        transf=@(x) x.^(1./wmat);
    else
        transf=@(x) x;
    end
    %transf=@(x) x;
    ss=transf(im(bnz, :));
    s0=transf(im(bz, :));
    
    [sqrt(sum(sum(ss.^2))), sqrt(sum(sum(s0.^2)))]
    sqrt(sum(sum(initf.^2)))
    
    rawsig=[length(bnz) 1 w h d];
    sig=[3 2 w h d];

    %ts=tsize_sym(sig);
    %res=zeros(ts, w*h*d);
    %ok=all(isreal(c) & isfinite(c), 1);
    %res(:, ok)=A \ c(:, ok);

    % calculate mask
    avg=reshape(sum(abs(data.I), 4), 1, w*h*d);
    threshold=0.10*max(avg);
    mask=avg>threshold;
    save(maskname, 'mask');

    
    % Write the results
    write_doubles(infilename, ss);
    write_doubles(s0filename, s0);
    write_doubles(amatname, A);

    parf=fopen([name, '.params'], 'w');
    
    fprintf(parf, '-dim %d\n', sig(1));
    fprintf(parf, '-order %d\n', sig(2));
    fprintf(parf, '-fdim %d\n', rawsig(1));
    fprintf(parf, '-forder %d\n', rawsig(2));
    fprintf(parf, '-width %d\n', sig(3));
    fprintf(parf, '-height %d\n', sig(4));
    fprintf(parf, '-depth %d\n', sig(5));
    fprintf(parf, '-amat %s\n', amatname);
    fprintf(parf, '-infile %s\n', infilename);
    fprintf(parf, '-s0file %s\n', s0filename);
    
    if ~isempty(initf)
        write_doubles(initfilename, initf);
        fprintf(parf, '-initfile %s\n', initfilename);
        fprintf(parf, '-initstyle file\n');
    end
    
    fclose(parf);
end

function res=kx(a)
    res=index_sym(3, a);
end
