/*
 * dticode/mex/sym2eigs_multi.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */

#include <string.h>
#include <stdlib.h>
#include <mex.h>
#include <stddef.h>
#include <math.h>
#include <lapack.h>

#include "../src/sym2multi3.h"
#include "../src/sym2multi2.h"
#include "../src/eig_qr.c"

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    double *vs;
    int i, ii, n, j;
    double *lp, *ep;
    double tmp[I_N];
    double evs[INS_N];
    int calc_evs=0;
    int dim=3;
    int res;
    
    if(nrhs!=1)
        mexErrMsgTxt("Invalid number of input arguments");
    if(nlhs!=2 && nlhs!=1)
        mexErrMsgTxt("Invalid number of output arguments");

    if(!mxIsDouble(prhs[0]))
        mexErrMsgTxt("Double matrix expected");
        
    n=mxGetN(prhs[0]);
    j=mxGetM(prhs[0]);
    vs=mxGetPr(prhs[0]);
    
    if(j==I2_N)
        dim=2;
    else if(j!=I_N)
        mexErrMsgTxt("Erroneous dimension");

    plhs[0]=mxCreateDoubleMatrix(dim, n, mxREAL);
    lp=mxGetPr(plhs[0]);
    if(nlhs==2){
        mwSize dims[3]={dim, dim, n};
        calc_evs=1;
        plhs[1]=mxCreateNumericArray(3, dims, mxDOUBLE_CLASS, mxREAL);
        ep=mxGetPr(plhs[1]);
    }
    
    if(dim==3){
        for(i=0; i<n; i++){
            ii=i*I_N;
            memcpy(tmp, vs+ii, sizeof(double)*I_N);
            res=eig_qr(evs, tmp, calc_evs);
            if(!res)
                mexErrMsgTxt("eig error");
            j=3*i;
            lp[j  ]=tmp[I_XX];
            lp[j+1]=tmp[I_YY];
            lp[j+2]=tmp[I_ZZ];
            if(calc_evs){
                j=INS_N*i;
                /*memcpy(ep+j, evs, sizeof(double)*INS_N);*/
                ep[j  ]=evs[INS_XX];
                ep[j+1]=evs[INS_XY];
                ep[j+2]=evs[INS_XZ];
                
                ep[j+3]=evs[INS_YX];
                ep[j+4]=evs[INS_YY];
                ep[j+5]=evs[INS_YZ];
                
                ep[j+6]=evs[INS_ZX];
                ep[j+7]=evs[INS_ZY];
                ep[j+8]=evs[INS_ZZ];
            }
        }
    }else{
        for(i=0; i<n; i++){
            ii=i*I2_N;
            res=eig2(evs, lp+2*i, vs+ii, calc_evs);
            if(!res)
                mexErrMsgTxt("eig error");
            if(calc_evs){
                j=INS2_N*i;
                /*memcpy(ep+j, evs, sizeof(double)*INS2_N);*/
                ep[j  ]=evs[INS2_XX];
                ep[j+1]=evs[INS2_XY];
                
                ep[j+2]=evs[INS2_YX];
                ep[j+3]=evs[INS2_YY];
            }
        }
    }
}
