/*
 * dticode/src/acc_support.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */


#include <stdlib.h>
#include <string.h>
#ifdef _OPENACC
#include <openacc.h>
#ifdef __PGI
#define __align__(X)
#define __location__(X)
#define CUDARTAPI
#endif
#include <cuda_runtime_api.h>
#endif

#include "acc_support.h"
#include "number.h"

void memcpy_dev(void * DEVPTR devptr, const void * DEVPTR devptr2, size_t n)
{
#ifndef _OPENACC
    memcpy(devptr, devptr2, n);
#else
    cudaMemcpy(devptr, devptr2, n, cudaMemcpyDeviceToDevice);
#endif
}

void memcpy_to_dev(void *devptr, const void *hostptr, size_t n)
{
#ifndef _OPENACC
    memcpy(devptr, hostptr, n);
#else
    cudaMemcpy(devptr, hostptr, n, cudaMemcpyHostToDevice);
#endif
}

void memcpy_from_dev(void *hostptr, const void *devptr, size_t n)
{
#ifndef _OPENACC
    memcpy(hostptr, devptr, n);
#else
    cudaMemcpy(hostptr, devptr, n, cudaMemcpyDeviceToHost);
#endif
}

void veccopy_to_dev(FNumber * DEVPTR devptr, const FNumber *src, size_t n)
{
    memcpy_to_dev(devptr, src, sizeof(FNumber)*n);
}

void veccopy_from_dev(FNumber *ptr, const FNumber * DEVPTR devsrc, size_t n)
{
    memcpy_from_dev(ptr, devsrc, sizeof(FNumber)*n);
}

FNumber *vector_to_dev(FNumber *ptr, size_t n, bool dofree)
{
#ifndef _OPENACC
    return ptr;
#else
    FNumber * DEVPTR devptr = acc_malloc(sizeof(FNumber)*n);
    if(devptr==NULL)
        die_err();
    veccopy_to_dev(devptr, ptr, n);
    if(dofree)
        free(ptr);
    return devptr;
#endif
}

FNumber *vector_from_dev(FNumber * DEVPTR devptr, size_t n, bool dofree)
{
#ifndef _OPENACC
    return devptr;
#else
    FNumber * ptr = malloc(sizeof(FNumber)*n);
    if(ptr==NULL)
        die_err();
    veccopy_from_dev(ptr, devptr, n);
    if(dofree)
        acc_free(devptr);
    return ptr;
#endif
}

