/*
 * dticode/src/combine.h
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */


#ifndef _COMBINE_H
#define _COMBINE_H

#include "cpock_accel.h"
#include "vector.h"
#include "splitvar.h"

typedef void FidParams;
typedef void RegParams;
typedef Vector FidPrimal;
typedef Vector RegExtraPrimal;
typedef Vector FidDual;
typedef Vector RegDual;

typedef SplitVar CombinedPrimal;
#define COMBINED_U(v) ((v)->p[0])
#define COMBINED_W(v) ((v)->p[1])

typedef SplitVar CombinedDual;
#define COMBINED_FID(v) ((v)->p[0])
#define COMBINED_REG(v) ((v)->p[1])

typedef enum{
    FidBaseshiftNone,
    FidBaseshiftSet,
    FidBaseshiftAdd,
    FidBaseshiftBregman,
} FidBaseshiftMode;

typedef enum{
    FidInitPrimalZero,
    FidInitPrimalBackproj,
    FidInitPrimalFile,
} FidInitPrimalStyle;

typedef enum{
    EstimateKNorm,
    EstimateGUniformConvexity,
    EstimateFCUniformConvexity,
} Estimate;

typedef void FidGres(FidPrimal *u, double tau, FidParams *p);
typedef void FidFCres(FidDual *phi, double sigma, FidParams *p);
/* FidKC has to add result to u! */
typedef void FidK(FidDual *phi, const FidPrimal *u, bool addto_phi, FidParams *p);
typedef void FidKC(FidPrimal *u, const FidDual *phi, bool addto_u, FidParams *p);
typedef void FidPDgap(FNumber *pdgap, const FidPrimal *x, const FidDual *y,
                      const FidPrimal *kcy, const FidDual *kx, 
                      FidPrimal *xwork, unsigned int iter,
                      FidParams *mp);
typedef FidPrimal *FidAllocPrimal(FidParams *mp);
typedef FidDual *FidAllocDual(FidParams *mp);
typedef void FidInitPrimal(FidPrimal *u, FidInitPrimalStyle style, FidParams *mp);
typedef void FidInitDual(FidDual *phi, const FidPrimal *u, FidParams *mp);
typedef FNumber FidBaseshift(FidPrimal *u, FidBaseshiftMode mode, FidParams *mp);
typedef FNumber FidEstimate(Estimate estim, const FidPrimal *u, FidParams *mp);
// FidGrad returns FALSE if only a subgradient
typedef bool FidGrad(FidPrimal *grad, const FidPrimal *u, FidParams *mp);

typedef void RegGres(RegExtraPrimal *w, double tau, RegParams *p);
typedef void RegFCres(RegDual *phi, double sigma, RegParams *p);
typedef void RegK(RegDual *phi, const CombinedPrimal *uw, bool addto_phi, RegParams *p);
typedef void RegKC(CombinedPrimal *uw, const RegDual *psi, 
                   bool addto_u, bool addto_w, RegParams *p);
typedef void RegPDgap(FNumber *pdgap, const CombinedPrimal *x, const FidDual *y,
                      const CombinedPrimal *kcy, const FidDual *kx, 
                      CombinedPrimal *xwork, unsigned int iter,
                      RegParams *mp);
typedef RegExtraPrimal *RegAllocPrimal(RegParams *mp);
typedef RegDual *RegAllocDual(RegParams *mp);
typedef void RegInitDualExtra(RegExtraPrimal *w, RegDual *psi, const FidPrimal *u, RegParams *mp);
typedef FNumber RegEstimate(Estimate estim, RegParams *mp);

typedef struct{
    FidK *k;
    FidKC *kc;
    FidFCres *fcres;
    FidGres *gres;
    FidPDgap *pdgap;
    FidAllocPrimal *alloc_primal;
    FidAllocDual *alloc_dual;
    FidInitPrimal *init_primal;
    FidInitDual *init_dual;
    FidBaseshift *baseshift;
    FidEstimate *estimate;
    FidGrad *grad;
} Fid;

typedef struct{
    RegK *k;
    RegKC *kc;
    RegFCres *fcres;
    RegGres *gres;
    RegPDgap *pdgap;
    RegAllocPrimal *alloc_primal;
    RegAllocDual *alloc_dual;
    RegInitDualExtra *init_dual_extra;
    RegEstimate *estimate;
} Reg;

extern void noop();

extern void combine_do_cpock(CPockParam *cpockp,
                             const Fid *fid, FidParams *fidp,
                             const Reg *reg, RegParams *regp,
                             FidInitPrimalStyle initstyle, 
                             const void *initparam,
                             CombinedPrimal **pp,
                             CombinedDual **dp);

/* internal */

typedef struct{
    ModelParams *fidp;
    ModelParams *regp;
    const Fid *fid;
    const Reg *reg;
} CombinedParams;

CombinedDual *combined_alloc_dual(CombinedParams *cp);
CombinedPrimal *combined_alloc_primal(CombinedParams *cp);
void combined_free_dual(CombinedDual *d, CombinedParams *cp);
void combined_free_primal(CombinedPrimal *d, CombinedParams *cp);

extern void combined_init(CombinedPrimal *p0, CombinedDual *d0, 
                          FidInitPrimalStyle initstyle, 
                          const void *initparam,
                          const Fid *fid, FidParams *fidp,
                          const Reg *reg, RegParams *regp);

extern FNumber combined_estimate(Estimate estim,
                                 const CombinedPrimal *p,
                                 const Fid *fid, FidParams *fidp,
                                 const Reg *reg, RegParams *regp);

extern void combined_pdgap(FNumber *pdgap, 
                           const CombinedPrimal *x, 
                           const CombinedDual *y,
                           const CombinedPrimal *kcy, 
                           const CombinedDual *kx, 
                           CombinedPrimal *xwork, 
                           unsigned int iter,
                           CombinedParams *cp);

extern CPockFn combined_fn;

#endif /* _COMBINE_H */
