/*
 * dticode/src/combine_nonlin.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <libtu/misc.h>
#include <libtu/output.h>
#include "combine_nonlin.h"
#include "cpock_nonlin.h"


static void combined_nl_k(CombinedDual *a, 
                          const CombinedPrimal *b,
                          CombinedNLParams *cp)
{
    cp->reg->k(COMBINED_REG(a), b, FALSE, cp->regp);
    cp->fid->nl_k(COMBINED_FID(a), COMBINED_U(b), FALSE, cp->fidp);
}

static void combined_nl_kag(CombinedDual *a, 
                             const CombinedPrimal *b0,
                             const CombinedPrimal *b,
                             CombinedNLParams *cp)
{
    cp->reg->k(COMBINED_REG(a), b, FALSE, cp->regp);
    cp->fid->nl_kag(COMBINED_FID(a), COMBINED_U(b0), COMBINED_U(b), FALSE, cp->fidp);
}

static void combined_nl_kc(CombinedPrimal *a, 
                           const CombinedPrimal *a0,
                           const CombinedDual *b,
                           CombinedNLParams *cp)
{
    cp->reg->kc(a, COMBINED_REG(b), FALSE, FALSE, cp->regp);
    cp->fid->nl_kc(COMBINED_U(a), COMBINED_U(a0), COMBINED_FID(b), TRUE, cp->fidp);
}

static void combined_nl_fcres(CombinedDual *a, 
                              FNumber sigma, 
                              CombinedNLParams *cp)
{
    cp->reg->fcres(COMBINED_REG(a), sigma, cp->regp);
    cp->fid->lin.fcres(COMBINED_FID(a), sigma, cp->fidp);
}


static void combined_nl_gres(CombinedPrimal *a, 
                          FNumber sigma, 
                          CombinedNLParams *cp)
{
    cp->reg->gres(COMBINED_W(a), sigma, cp->regp);
    cp->fid->lin.gres(COMBINED_U(a), sigma, cp->fidp);
}

CombinedDual *combined_nl_alloc_dual(CombinedNLParams *cp)
{
    CombinedDual *d=alloc_splitvar(2);
    
    COMBINED_FID(d)=cp->fid->lin.alloc_dual(cp->fidp);
    COMBINED_REG(d)=cp->reg->alloc_dual(cp->regp);
    
    return d;
}

CombinedPrimal *combined_nl_alloc_primal(CombinedNLParams *cp)
{
    CombinedPrimal *d=alloc_splitvar(2);
        
    COMBINED_U(d)=cp->fid->lin.alloc_primal(cp->fidp);
    COMBINED_W(d)=cp->reg->alloc_primal(cp->regp);
    
    return d;
}

FNumber combined_nl_estimate(Estimate estim, const CombinedPrimal *p,
                             CombinedNLParams *cp)
{
    return combined_estimate(estim, p, &cp->fid->lin, cp->fidp, cp->reg, cp->regp);
}

CPockNLFn combined_nl_fn={
    (CPockNLK*)combined_nl_k,
    (CPockNLKAG*)combined_nl_kag,
    (CPockNLKC*)combined_nl_kc,
    (CPockFCres*)combined_nl_fcres,
    (CPockGres*)combined_nl_gres,
    (CPockNLEstimate*)combined_nl_estimate,
    //(CPockPDgap*)combined_nl_pdgap,
};


FidPrimal *combine_do_cpock_nonlin(CPockParam *cpockp,
                                   const FidNL *fid, FidParams *fidp,
                                   const Reg *reg, RegParams *regp,
                                   FidInitPrimalStyle initstyle,
                                   const void *initparam)
{
    CombinedNLParams cp;
    CombinedPrimal *p, *p0, *ptmp;
    CombinedDual *d, *d0, *dtmp;
    FidPrimal *res;
    
    cp.reg=reg;
    cp.regp=regp;
    cp.fid=fid;
    cp.fidp=fidp;
    
    // Alloc
    p=combined_nl_alloc_primal(&cp);
    p0=combined_nl_alloc_primal(&cp);
    ptmp=combined_nl_alloc_primal(&cp);
    d=combined_nl_alloc_dual(&cp);
    d0=combined_nl_alloc_dual(&cp);
    dtmp=combined_nl_alloc_dual(&cp);

    // Init
    combined_init(p0, d0, initstyle, initparam, &(fid->lin), fidp, reg, regp);
    
    cpockp->knorm=combined_estimate(EstimateKNorm, p0, &(fid->lin), fidp, reg, regp);
    cpockp->gamma=fn_max(combined_estimate(EstimateGUniformConvexity, p0, &(fid->lin), fidp, reg, regp),
                         combined_estimate(EstimateFCUniformConvexity, p0, &(fid->lin), fidp, reg, regp));

    // Run the algorithm
    cpock_nonlin((Vector*)p, (Vector*)d, (Vector*)p0, 
                 (Vector*)d0, (Vector*)ptmp,  (Vector*)dtmp, 
                 cpockp, &combined_nl_fn, &cp);
    
    // Pick result, and prevent it from being unallocated
    res=COMBINED_U(p);
    COMBINED_U(p)=NULL;

    destroy_obj((Obj*)p);
    destroy_obj((Obj*)p0);
    destroy_obj((Obj*)ptmp);
    destroy_obj((Obj*)d);
    destroy_obj((Obj*)d0);
    destroy_obj((Obj*)dtmp);
    
    return res;
}

