/*
 * dticode/src/cpock_accel.h
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */


#ifndef _CPOCK_ACCEL_H
#define _CPOCK_ACCEL_H

#include "libtu/types.h"
#include "number.h"
#include "vector.h"

/* Just for readability */
typedef Vector Primal;
typedef Vector Dual;
typedef void ModelParams;

typedef enum{
    NLCPockExact,
    NLCPockLin,
    NLCPockInterp
} NLCPockMode;

/* Functions */
typedef struct{
    FNumber sigma;
    FNumber tau;
    FNumber gamma;
    FNumber rho;
    FNumber theta;
    FNumber knorm;
    unsigned int maxiter;
    bool verbose;
    int verbose_granularity;
    bool calculate_pdgap0;
    NLCPockMode mode;
} CPockParam;

typedef void CPockK(Dual *a, const Primal *b, ModelParams *mp);
typedef void CPockKC(Primal *a, const Dual *b, ModelParams *mp);
typedef void CPockFCres(Dual *a, FNumber sigma, ModelParams *mp);
typedef void CPockGres(Primal *a, FNumber sigma, ModelParams *mp);
typedef void CPockPDgap(FNumber *pdgap, const Primal *x, const Dual *y,
                        const Primal *kcy, const Dual *kx, 
                        Primal *xwork, unsigned int iter, 
                        ModelParams *mp);

typedef struct{
    CPockK *k;
    CPockKC *kc;
    CPockFCres *fcres;
    CPockGres *gres;
    CPockPDgap *pdgap;
} CPockFn;



extern void cpock_accel(Primal *xres, Dual *yres, Primal *x, Dual *y, 
                        Primal *xtmp, Dual *ytmp,
                        const CPockParam *p, 
                        const CPockFn *fn,
                        ModelParams *mp);

#endif /* _CPOCK_ACCEL_H */
