/*
 * dticode/src/denoise.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <libtu/util.h>
#include <libtu/output.h>
#include <libtu/optparser.h>

#include "parallel.h"
#include "cpock_accel.h"
#include "tensor.h"
#include "reg.h"
#include "fid.h"
#include "linoper_id.h"
#include "timing.h"
#include "combine.h"
#include "vector.h"
#include "method.h"


static OptParserOpt opts[]={
    QUICK_OPT('R', "reg",     "Regulariser: tv|tgv2|none,alpha,beta"),
    QUICK_OPT('c', "colour",  "Colour: gray|rgb"),
    QUICK_OPT('F', "fid",     "Fidelity: l2|l1|l2alt"),
    QUICK_OPT('f', "infile",  "Input file (raw doubles)"),
    METHOD_OPTS,
    END_OPTPARSEROPTS
};

int main(int argc, char *argv[])
{
    MethodParam mp=DEFAULT_METHOD_PARAM;
    
    mp.sig.dim=2;
    mp.sig.order=0;
    mp.sig.symmetric=0;
    
    const char *regspec="tv,1";
    const char *fidname="l2";
    const char *infile=NULL;
    const char *colour="gray";
    TensorField *f=NULL;
    int opt;
    FidP fidp;
    Fid *fid;
    RegP regp;
    Reg *reg;
    LinOper *op;

    parallel_init1(&argc, &argv);
    
    libtu_init(argv[0]);
    optparser_init(argc, argv, OPTP_MIDLONG, opts);
    
    while((opt=optparser_get_opt())){
        switch(opt){
            case OPT_ID('R'):
                regspec=optparser_get_arg();
                break;
            case OPT_ID('F'):
                fidname=optparser_get_arg();
                break;
            case OPT_ID('c'):
                colour=optparser_get_arg();
                break;
            case OPT_ID('f'):
                infile=optparser_get_arg();
                break;
            default:
                if(!method_do_opt(&mp, opt, opts)){
                    optparser_print_error();
                    return 1;
                }
                break;
        }
    }
    
    if(infile==NULL)
        die("infile missing\n");

    if(strcmp(colour, "rgb")==0){
        mp.sig.dim=3;
        mp.sig.order=1;
        mp.sig.d=1;
    }else if(strcmp(colour, "gray")==0){
        mp.sig.dim=2;
        mp.sig.order=0;
        mp.sig.d=1;
    }else{
        die("Invalid colour space: %s\n", colour);
    }

    method_params_sanity_check(&mp);

    parallel_init2();

    f=load_tensorfield(&mp.sig, infile);

    op=(LinOper*)alloc_linoper_id(&mp.sig);
    
    init_fidelity(&fid, &fidp, fidname, op, f, f, f);

    init_regulariser(&reg, &regp, regspec, &mp.sig, mp.diffalg);
    
    method_do(&mp, fid, &fidp, reg, &regp);
    
    /*
    if(fullsave){
        char *nm;
        libtu_asprintf(&nm, "%s.primal", outfile);
        vector_save((Vector*)pres, nm);
        free(nm);
        libtu_asprintf(&nm, "%s.dual", outfile);
        vector_save((Vector*)dres, nm);
        free(nm);
    }*/

    parallel_finalise();

    return 0;
}
