/*
 * dticode/src/diff_rules.h
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */


#ifndef _DIFF_RULES_H
#define _DIFF_RULES_H

#include <stdlib.h>

#include "tensor.h"
#include "number.h"
#include "sparse.h"
#include "acc_support.h"
#include "diff_workers.h"

#define ANY (-1)

#define CATX(X, Y) X ## _ ## Y
#define CAT(X, Y) CATX(X, Y)

/* forward-differences with zero boundary condition */
#define diffrule_f_0_First(a, aprev, anext)  (anext-a)
#define diffrule_f_0_Middle(a, aprev, anext) (anext-a)
#define diffrule_f_0_Last(a, aprev, anext)   (-a)

/* conjugate of forward-differences with zero boundary condition */
#define diffrule_reverse_f_0_First(a, aprev, anext)  (-a)
#define diffrule_reverse_f_0_Middle(a, aprev, anext) (aprev-a)
#define diffrule_reverse_f_0_Last(a, aprev, anext)   (aprev-a)

/* backward-differences with zero boundary condition */
#define diffrule_b_0_First(a, aprev, anext)  (a)
#define diffrule_b_0_Middle(a, aprev, anext) (a-aprev)
#define diffrule_b_0_Last(a, aprev, anext)   (a-aprev)

/* conjugate of backward-differences with zero boundary condition */
#define diffrule_reverse_b_0_First(a, aprev, anext)  (a-anext)
#define diffrule_reverse_b_0_Middle(a, aprev, anext) (a-anext)
#define diffrule_reverse_b_0_Last(a, aprev, anext)   (a)

/* central-differences with zero boundary condition */
#define diffrule_c_0_First(a, aprev, anext)  (-anext*0.5)
#define diffrule_c_0_Middle(a, aprev, anext) ((aprev-anext)*0.5)
#define diffrule_c_0_Last(a, aprev, anext)   (aprev*0.5)

/* conjugate of central-differences with zero boundary condition */
/* This is just the negation of diffrule_c_0 */
#define diffrule_reverse_c_0_First(a, aprev, anext)  (-(-anext)*0.5)
#define diffrule_reverse_c_0_Middle(a, aprev, anext) (-(aprev-anext)*0.5)
#define diffrule_reverse_c_0_Last(a, aprev, anext)   (-aprev*0.5)

/* Size hard-coding optimization hack defaults */

#ifndef VOLUME_SIZE_NAME
#define VOLUME_SIZE_NAME generic
#endif

#endif /* _DIFF_RULES_H */
