/*
 * dticode/src/diff_worker_template.h
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */

/* Hard-coding optimisation helper definitions. */
#if !defined(WIDTH) || WIDTH==ANY
#define _WIDTH w
#else
#define _WIDTH WIDTH
#endif

#if !defined(HEIGHT) || HEIGHT==ANY
#define _HEIGHT h
#else
#define _HEIGHT HEIGHT
#endif

#if !defined(DEPTH) || DEPTH==ANY
#define _DEPTH d
#else
#define _DEPTH DEPTH
#endif

#ifndef _DIM
#define _DIM dim
#else
#define _DIM DIM
#endif

#ifndef _RESDIM
#define _RESDIM resdim
#else
#define _RESDIM RESDIM
#endif

#if !defined(XRES) || XRES==ANY
#define _XRES xres
#else
#define _XRES XRES
#endif

#if !defined(YRES) || YRES==ANY
#define _YRES yres
#else
#define _YRES YRES
#endif

#if !defined(ZRES) || ZRES==ANY
#define _ZRES zres
#else
#define _ZRES ZRES
#endif

#if !defined(ALPHA) || ALPHA==ANY
#define _ALPHA alpha
#else
#define _ALPHA ALPHA
#endif

#if !defined(BETA) || BETA==ANY
#define _BETA beta
#else
#define _BETA BETA
#endif

#if _ALPHA==0
    #if _BETA==1
        #define DIFF_ASSIGN(X, Y) X=Y
    #else
        #define DIFF_ASSIGN(X, Y) X=(_BETA)*(Y)
    #endif
#else
    #define DIFF_ASSIGN(X, Y) X=(_ALPHA)*(X)+(_BETA)*(Y)
#endif

#if !defined(CALCDIFF1) && !defined(CALCDIFF2) && !defined(CALCDIFF3)
#define USE_SPARSEMATRIX 1
#define SPARSE_MMULT(resdim, res, dim, off, vec, vecp) \
    { \
        _Pragma("acc loop seq") \
        for(size_t _i=0; _i<dim; _i++){ \
            const FNumber _val=vec(_i, vecp); \
            const size_t _end2=(pm_jc+off)[_i+1]; \
            _Pragma("acc loop seq") \
            for(size_t _j=(pm_jc+off)[_i]; _j<_end2; _j++){ \
                const size_t _k=pm_ir[_j]; \
                res(_k)+=pm_pr[_j]*_val; \
            } \
        } \
    }

#define VECZERO(p, n) \
    { \
        for(size_t _k=0; _k<n; _k++) \
            p(_k)=0; \
    } 


#define CALCDIFF1(resthis, x, w, vecx) \
    { \
        VECZERO(resthis, _RESDIM); \
        SPLIT_OPTIM(x, w, SPARSE_MMULT, _RESDIM, resthis, _DIM, 0, vecx); \
    }

#define CALCDIFF2(resthis, x, y, w, h, vecx, vecy) \
    { \
        VECZERO(resthis, _RESDIM); \
        SPLIT_OPTIM(x, w, SPARSE_MMULT, _RESDIM, resthis, _DIM, 0, vecx); \
        SPLIT_OPTIM(y, h, SPARSE_MMULT, _RESDIM, resthis, _DIM, _DIM, vecy); \
    }

#define CALCDIFF3(resthis, x, y, z, w, h, d, vecx, vecy, vecz) \
    { \
        VECZERO(resthis, _RESDIM); \
        SPLIT_OPTIM(x, w, SPARSE_MMULT, _RESDIM, resthis, _DIM, 0, vecx); \
        SPLIT_OPTIM(y, h, SPARSE_MMULT, _RESDIM, resthis, _DIM, _DIM, vecy); \
        SPLIT_OPTIM(z, d, SPARSE_MMULT, _RESDIM, resthis, _DIM, 2*_DIM, vecz); \
    }

#endif /* defined(CALCDIFF?) */

#define RESTHIS(X) res[resoff+(X)]
#define ATHIS(X) a[aoff+(X)]

#define DIFFRULEX(i, where) (xres_div* CAT(DIFFRULE, where) (ATHIS(i), ATHIS(i-xoff), ATHIS(xoff+i)))
#define DIFFRULEY(i, where) (yres_div* CAT(DIFFRULE, where) (ATHIS(i), ATHIS(i-yoff), ATHIS(yoff+i)))
#define DIFFRULEZ(i, where) (zres_div* CAT(DIFFRULE, where) (ATHIS(i), ATHIS(i-zoff), ATHIS(zoff+i)))
#define DIFFRULE0(i, where) 0

/* Force compile-time decidable coordinate comparisons in DIFFRULE */
#define SPLIT_OPTIM(COORD, N, ACT, ...) \
    if(COORD==0){ \
        ACT(__VA_ARGS__, First); \
    }else if(COORD==N-1){ \
        ACT(__VA_ARGS__, Last); \
    }else{ \
        ACT(__VA_ARGS__, Middle); \
    }
        

static void WORKER(size_t resdim, size_t dim,
                   size_t w, size_t h, size_t d,
                   FNumber xres, FNumber yres, FNumber zres,
                   FNumber * restrict DEVPTR res,
                   const FNumber * restrict DEVPTR a,
                   FNumber alpha, FNumber beta,
                   const SparseMatrix *pm)
{
    const size_t xoff=_DIM;
    const size_t yoff=_WIDTH*_DIM;
    const size_t zoff=_WIDTH*_HEIGHT*_DIM;
    const size_t rxoff=_RESDIM;
    const size_t ryoff=_WIDTH*_RESDIM;
    const size_t rzoff=_WIDTH*_HEIGHT*_RESDIM;
    const FNumber xres_div=FNUMBER(1.0)/_YRES;
    const FNumber yres_div=FNUMBER(1.0)/_XRES;
    const FNumber zres_div=FNUMBER(1.0)/_ZRES;
#if USE_SPARSEMATRIX
    const size_t irsize=pm->irsize;
    const size_t pmcols=pm->n;
    const FNumber * const restrict pm_pr=pm->pr;
    const size_t * const restrict pm_ir=pm->ir;
    const size_t * const restrict pm_jc=pm->jc;
    #pragma acc data copyin(pm_jc[0:pmcols+1], pm_ir[0:irsize], pm_pr[0:irsize])
#endif


#if VOLUME_DIM==1
    #pragma omp parallel 
    #pragma acc kernels deviceptr(res, a)
    {
        #pragma omp for
        #pragma acc loop gang, vector, independent
        for(size_t x=0; x<_WIDTH; x++){
            const size_t resoff=x*rxoff;
            const size_t aoff=x*xoff;
            #pragma acc cache(a[aoff:_DIM], \
                              a[aoff+xoff:_DIM], \
                              a[aoff-xoff:_DIM], \
                              res[resoff:_RESDIM])
            CALCDIFF1(RESTHIS, x, _WIDTH, DIFFRULEX);
        }
    }
#elif VOLUME_DIM==2
    {
        #pragma omp parallel 
        #pragma acc kernels deviceptr(res, a)
        {
            #pragma omp for
            #pragma acc loop gang, vector, collapse(2), independent
            for(size_t y=0; y<_HEIGHT; y++){
                for(size_t x=0; x<_WIDTH; x++){
                    const size_t resoff = y*ryoff+x*rxoff;
                    const size_t aoff = y*yoff+x*xoff;
                    #pragma acc cache(a[aoff:_DIM], \
                                      a[aoff+xoff:_DIM], \
                                      a[aoff-xoff:_DIM], \
                                      a[aoff+yoff:_DIM], \
                                      a[aoff-yoff:_DIM], \
                                      res[resoff:_RESDIM])
                    CALCDIFF2(RESTHIS, x, y, _WIDTH, _HEIGHT, 
                              DIFFRULEX, DIFFRULEY);
                }
            }
        }
    }
#elif VOLUME_DIM==3
    if(d>1){
        #pragma omp parallel
        #pragma acc kernels deviceptr(res, a)
        {
            #pragma omp for
            //#pragma acc loop collapse(3), independent
            #pragma acc loop gang, vector, collapse(3), independent
            for(size_t z=0; z<_DEPTH; z++){
                for(size_t y=0; y<_HEIGHT; y++){
                    for(size_t x=0; x<_WIDTH; x++){
                        const size_t resoff=z*rzoff+y*ryoff+x*rxoff;
                        const size_t aoff=z*zoff+y*yoff+x*xoff;
                        #pragma acc cache(a[aoff:_DIM], \
                                          a[aoff+xoff:_DIM], \
                                          a[aoff-xoff:_DIM], \
                                          a[aoff+yoff:_DIM], \
                                          a[aoff-yoff:_DIM], \
                                          a[aoff+zoff:_DIM], \
                                          a[aoff-zoff:_DIM], \
                                          res[resoff:_RESDIM])
                        CALCDIFF3(RESTHIS, x, y, z, 
                                  _WIDTH, _HEIGHT, _DEPTH,
                                  DIFFRULEX, DIFFRULEY, DIFFRULEZ);
                    }
                }
            }
        }
    }else{
        #pragma omp parallel 
        #pragma acc kernels deviceptr(res, a)
        {
            #pragma omp for
            #pragma acc loop gang, vector, collapse(2), independent
            for(size_t y=0; y<h; y++){
                for(size_t x=0; x<w; x++){
                    const size_t resoff=y*ryoff+x*rxoff;
                    const size_t aoff=y*yoff+x*xoff;
                    #pragma acc cache(a[aoff:_DIM], \
                                      a[aoff+xoff:_DIM], \
                                      a[aoff-xoff:_DIM], \
                                      a[aoff+yoff:_DIM], \
                                      a[aoff-yoff:_DIM], \
                                      res[resoff:_RESDIM])
                    CALCDIFF3(RESTHIS, x, y, 0,
                              _WIDTH, _HEIGHT, 1,
                              DIFFRULEX, DIFFRULEY, DIFFRULE0);
                }
            }
        }
    }
#else        
    #error "Differentation implemented only for volumes dimension <= 3"
#endif
}

#undef ATHIS
#undef RESTHIS

#undef _WIDTH
#undef _HEIGHT
#undef _DEPTH
#undef _DIM
#undef _RESDIM
#undef _XRES
#undef _YRES
#undef _ZRES
#undef _ALPHA
#undef _BETA

#undef DIFF_UPDATE

#undef SPLIT_OPTIM

#undef DIFFRULEX
#undef DIFFRULEY
#undef DIFFRULEZ

#undef DIFFRULE
#undef WORKER

#undef CALCDIFF1
#undef CALCDIFF2
#undef CALCDIFF3

#undef USE_SPARSEMATRIX
