/*
 * dticode/src/diffoper.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <libtu/types.h>
#include <libtu/objp.h>
#include "tensor.h"
#include "vector.h"
#include "number.h"
#include "diffoper.h"

// calculate y=A(x)
void diffoper_calculate(DOper *op,
                        DOperCod *y, 
                        const DOperDom *x,
                        bool addto_y)
{
    CALL_DYN(diffoper_calculate, op, (op, y, x, addto_y));
}

// calculate x in inverse A(y)
void diffoper_backproj(DOper *op,
                       DOperDom *x, 
                       const DOperCod *y)
{
    CALL_DYN(diffoper_backproj, op, (op, x, y));

}                    

// calculate y=[grad A(x0)] x
void diffoper_apply_grad(DOper *op,
                         DOperCod *y, 
                         const DOperDom *x0, 
                         const DOperDom *x, 
                         bool addto_y)
{
    CALL_DYN(diffoper_apply_grad, op, (op, y, x0, x, addto_y));
}                      
                             
// calculate x=[grad A(x0)]^* y
void diffoper_apply_gradc(DOper *op,
                          DOperDom *x, 
                          const DOperDom *x0, 
                          const DOperCod *y, 
                          bool addto_x)
{
    CALL_DYN(diffoper_apply_gradc, op, (op, x, x0, y, addto_x));
}

static void diffoper_projdom_default(DOper *op, DOperDom *x)
{
}

void diffoper_projdom(DOper *op, DOperDom *x)
{
    CALL_DYN(diffoper_projdom, op, (op, x));
}

// Calculate the support function of any constraint set on the codomain
// variables (in order to calculate a meaningful pseudo-duality gap, when
// the duality gap would be useless). In essence this is the norm of the
// domain.
FNumber diffoper_constraint_scale(DOper *op, const DOperDom *x)
{
    FNumber ret=0;
    
    CALL_DYN_RET(ret, FNumber, diffoper_constraint_scale, op, (op, x));
    
    return ret;
}


// Calculate the conjugate of the (unit) constraint. In essence this
// is the dual norm.
FNumber diffoper_constraint_conjugate(DOper *op, DOperDom *x)
{
    FNumber ret=0;
    
    CALL_DYN_RET(ret, FNumber, diffoper_constraint_conjugate, op, (op, x));
    
    return ret;
}

FNumber diffoper_constraint_scale_default(DOper *op, const DOperDom *x)
{
    // L2 norm assumed
    return vector_norm(x);
}

FNumber diffoper_constraint_conjugate_default(DOper *op, const DOperDom *x)
{
    // L2 norm assumed
    return vector_norm(x);
}

// Estimate |grad G(x)|
FNumber diffoper_estimate_grad_norm(DOper *op, const DOperDom *x)
{
    FNumber ret=0;
    
    CALL_DYN_RET(ret, FNumber, diffoper_estimate_grad_norm, op, (op, x));
    
    return ret;
}

// Allocate a variable in the domain.
DOperDom *diffoper_alloc_dom(DOper *op)
{
    DOperDom *ret=NULL;
    
    CALL_DYN_RET(ret, DOperDom*, diffoper_alloc_dom, op, (op));
    
    return ret;
}

// Allocate a variable in the codomain.
DOperCod *diffoper_alloc_cod(DOper *op)
{
    DOperCod *ret=NULL;
    
    CALL_DYN_RET(ret, DOperCod*, diffoper_alloc_cod, op, (op));
    
    return ret;
}

bool diffoper_init(DOper *a)
{
    return TRUE;
}

void diffoper_deinit(DOper *a)
{
}

static DynFunTab diffoper_dynfuntab[]={
    {diffoper_projdom, diffoper_projdom_default},
    {(DynFun*)diffoper_constraint_scale, (DynFun*)diffoper_constraint_scale_default},
    {(DynFun*)diffoper_constraint_conjugate, (DynFun*)diffoper_constraint_conjugate_default},
    END_DYNFUNTAB,
};

IMPLCLASS(DOper, Obj, diffoper_deinit, diffoper_dynfuntab);
