/*
 * dticode/src/diffoper.h
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */


#ifndef _DIFFOPER_H
#define _DIFFOPER_H

#include <libtu/types.h>
#include <libtu/obj.h>
#include "tensor.h"
#include "vector.h"

INTRCLASS(DOper);

DECLCLASS(DOper){
    Obj obj;
};


typedef Vector DOperDom;
typedef Vector DOperCod;

// calculate y=A(x)
extern void diffoper_calculate(DOper *op,
                               DOperCod *y, 
                               const DOperDom *x,
                               bool addto_y);

// calculate x in inverse A(y)
extern void diffoper_backproj(DOper *op,
                              DOperDom *x, 
                              const DOperCod *y);

// calculate y=[grad A(x0)] x
extern void diffoper_apply_grad(DOper *op,
                                DOperCod *y, 
                                const DOperDom *x0, 
                                const DOperDom *x, 
                                bool addto_y);
                             
// calculate x=[grad A(x0)]^* y
extern void diffoper_apply_gradc(DOper *op,
                                 DOperDom *x, 
                                 const DOperDom *x0, 
                                 const DOperCod *y, 
                                 bool addto_x);

extern void diffoper_projdom(DOper *op, DOperDom *x);

// Allocate a variable in the domain.
extern DOperDom *diffoper_alloc_dom(DOper *op);

// Allocate a variable in the codomain.
extern DOperCod *diffoper_alloc_cod(DOper *op);

// Calculate the support function of any constraint set on the codomain
// variables (in order to calculate a meaningful pseudo-duality gap, when
// the duality gap would be useless). In essence this is the norm of the
// domain.
extern FNumber diffoper_constraint_scale(DOper *op, const DOperDom *x);

// Calculate the conjugate of the (unit) constraint. In essence this
// is the dual norm.
extern FNumber diffoper_constraint_conjugate(DOper *op, DOperDom *x);

// Estimate |grad G(x)|
extern FNumber diffoper_estimate_grad_norm(DOper *op, const DOperDom *x);

extern bool diffoper_init(DOper *op);
extern void diffoper_deinit(DOper *op);

#endif /* _DIFFOPER_H */
