/*
 * dticode/src/eig_qr.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <stddef.h>

#include "number.h"
#include "sym2multi3.h"
#include "sym2multi2.h"
#include "varmatrix.h"
#include "misc.h"
#include "number.h"

/*
 * 2D
 */
#define DO_EIG2_LOWLEVEL(l1, l2, v1, v2, w, v, u) \
    { \
        const FNumber _u2=u*u; \
        const FNumber _d=(w-v)/2; \
        const FNumber _sq=fn_sqrt(_d*_d+_u2); \
        l1=v+_d-_sq; \
        l2=v+_d+_sq; \
        const FNumber _wml1=w-l1; \
        const FNumber _no=fn_sqrt(_wml1*_wml1+_u2); \
        /* no!=0 when called! */ \
        v1=-u/_no; \
        v2=_wml1/_no; \
    }

#define DO_EIG2(L1, L2, Q, A, calc_evs) \
    if(fn_abs(A(I2_XY)) <= fn_sqrt(fn_abs(A(I2_XX))*fn_abs(A(I2_YY)))*FNUMBER_EPS){ \
        L1=A(I2_XX); L2=A(I2_YY); \
        if(calc_evs){ \
            Q(INS2_XX)=1;  Q(INS2_YX)=0; \
            Q(INS2_XY)=0;  Q(INS2_YY)=1; \
        } \
    }else{ \
        FNumber v1, v2; \
        DO_EIG2_LOWLEVEL(L1, L2, v1, v2, A(I2_XX), A(I2_YY), A(I2_XY)); \
        if(calc_evs){ \
            Q(INS2_XX)=v1;  Q(INS2_YX)=v2; \
            Q(INS2_XY)=-v2; Q(INS2_YY)=v1; \
        } \
    } \

/* input: a, symmetric;
 * output: q, eigenvectors
 *         l, eigenvalues
 */
int eig2(FNumber q[INS2_N], FNumber l[2], const FNumber a[I2_N], int calc_evs)
{
    #define A(I) a[I]
    #define Q(I) q[I]
    #define L1 l[0]
    #define L2 l[1]
    DO_EIG2(L1, L2, Q, A, calc_evs);
    #undef L2
    #undef L1
    #undef Q
    #undef A
    
    return 1;
}


/*
 * 3D QR
 */

#define GIVENS(c, s, a, b) \
    { \
        FNumber _r, _d; \
        \
        if(b==0){ \
            c=1; \
            s=0; \
        }else{ \
            if(fn_abs(b)>fn_abs(a)){ \
                _r=-a/b; \
                _d=fn_sqrt(1+_r*_r); \
                s=1/_d; \
                c=_r/_d; \
            }else{ \
                _r=-b/a; \
                _d=fn_sqrt(1+_r*_r); \
                c=1/_d; \
                s=_r/_d; \
            } \
        } \
    }


/* input: a, symmetric;
 * output: ev, eigenvectors,
 *         diagonal of a, eigenvalues.
 */

int eig_qr(FNumber ev[INS_N], FNumber a[I_N], int calc_evs)
{
    int converged=0;
    
    #define A(I) a[I]
    #define EV(I) ev[I]
    #include "eig_qr_core.h"
    #undef EV
    #undef A
    
    return converged;
}

