/*
 * dticode/src/eig_qr_core.h
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */

#define EIG_QR_MAX_ITS 30
/* Parameters: 
 *   #define A(I)       Input: the source matrix A 
 *                      (symmetric 3x3, indices I_*).
 *                      Output: eigenvalues on diagonal.
 *   #define EV(I)      Output: eigenvectors of a (if calc_evs!=0),
 *                      (3x3 matrix, indices INS_*).
 *   int calc_evs       If non-zero, calculate eigenvectors.
 */
{
    const FNumber smlnum=FNUMBER_SFMIN/FNUMBER_PREC;
    const FNumber bignum=FNUMBER(1.0)/smlnum;
    const FNumber rmin=fn_sqrt(smlnum);
    const FNumber rmax=fn_sqrt(bignum);
    
    /* Scale the matrix A to safe range.
     */
    FNumber sigma=1; 
    int iscale=0;
    {
        //const FNumber anrm=vecmaxabs(a, I_N);
        FNumber anrm=fn_max(A(I_XX), 
                     fn_max(A(I_XY), 
                     fn_max(A(I_YY), 
                     fn_max(A(I_XZ), 
                     fn_max(A(I_YZ), 
                            A(I_ZZ))))));
        
        if(anrm>0 && anrm<rmin){
            iscale=1;
            sigma=rmin/anrm;
        }else if(anrm>rmax){
            iscale=1;
            sigma=rmax/anrm;
        }
        /*vecscale(amat, I_N, sigma); */
        A(I_XX)*=sigma;
        A(I_XY)*=sigma;
        A(I_YY)*=sigma;
        A(I_XZ)*=sigma;
        A(I_YZ)*=sigma;
        A(I_ZZ)*=sigma;
    }
    
    /* Tridiagonalise the symmetric 3x3 matrix A as A=E^t T E by
     * a Givens (row) transformation E.
     * A is overwritten by the tridiagonalised matrix T,
     * while E is written into ev.
     */
    {
        DEFINE_VARMATRIX_SYM3(qtaqmat);
        #define QtAQ(I) VAR_VARMATRIX(qtaqmat, I)
        
        FNumber c, s;
        GIVENS(c, s, A(I_YX), A(I_ZX));
        const FNumber sc=s*c;
        const FNumber c2=c*c;
        const FNumber s2=s*s;
        
        /* Q=[1 0 0; 0 c s; 0 -s c]; Qt=[1 0 0; 0 c -s; 0 s c]; */
        /*
        QtA[I_XX]=A(I_XX);
        QtA[I_XY]=A(I_XY);
        QtA[I_XZ]=A(I_XZ);
        QtA[I_YX]=c*A(I_YX)-s*A(I_ZX);
        QtA[I_YY]=c*A(I_YY)-s*A(I_ZY);
        QtA[I_YZ]=c*A(I_YZ)-s*A(I_ZZ);
        QtA[I_ZX]=s*A(I_YX)+c*A(I_ZX);
        QtA[I_ZY]=s*A(I_YY)+c*A(I_ZY);
        QtA[I_ZZ]=s*A(I_YZ)+c*A(I_ZZ);
        
        QtAQ[I_XX]=QtA[I_XX];
        QtAQ[I_XY]=c*QtA[I_XY]-s*QtA[I_XZ];
        QtAQ[I_XZ]=s*QtA[I_XY]+c*QtA[I_XZ];
        QtAQ[I_YX]=QtA[I_YX];
        QtAQ[I_YY]=c*QtA[I_YY]-s*QtA[I_YZ];
        QtAQ[I_YZ]=s*QtA[I_YY]+c*QtA[I_YZ];
        QtAQ[I_ZX]=QtA[I_ZX];
        QtAQ[I_ZY]=c*QtA[I_ZY]-s*QtA[I_ZZ];
        QtAQ[I_ZZ]=s*QtA[I_ZY]+c*QtA[I_ZZ];
        ==>
        */

        QtAQ(I_XX)=A(I_XX);
        QtAQ(I_XY)=c*A(I_XY)-s*A(I_XZ);
        QtAQ(I_XZ)=s*A(I_XY)+c*A(I_XZ);
        /*QtAQ(I_YX)=c*A(I_YX)-s*A(I_ZX);*/
        QtAQ(I_YY)=c2*A(I_YY)-sc*A(I_ZY)-sc*A(I_YZ)+s2*A(I_ZZ);
        QtAQ(I_YZ)=sc*A(I_YY)-s2*A(I_ZY)+c2*A(I_YZ)-sc*A(I_ZZ);
        /*QtAQ(I_ZX)=s*A(I_YX)+c*A(I_ZX);*/
        /*QtAQ(I_ZY)=sc*A(I_YY)+c2*A(I_ZY)-s2*A(I_YZ)-sc*A(I_ZZ);*/
        QtAQ(I_ZZ)=s2*A(I_YY)+sc*A(I_ZY)+sc*A(I_YZ)+c2*A(I_ZZ);
        
        /* pgcc can't handle veccopy... */
        /*veccopy(a, QtAQ, I_N);*/
        /*#pragma acc loop vector
        for(size_t i=0; i<I_N; i++)
            A(i)=QtAQ(i);*/
        A(I_XX)=QtAQ(I_XX);
        A(I_XY)=QtAQ(I_XY);
        A(I_YY)=QtAQ(I_YY);
        A(I_XZ)=QtAQ(I_XZ);
        A(I_YZ)=QtAQ(I_YZ);
        A(I_ZZ)=QtAQ(I_ZZ);
        
        EV(INS_XX)=1; 
        EV(INS_XY)=0;
        EV(INS_XZ)=0;
        EV(INS_YX)=0; 
        EV(INS_YY)=c;
        EV(INS_YZ)=s;
        EV(INS_ZX)=0; 
        EV(INS_ZY)=-s;
        EV(INS_ZZ)=c;
        
        #undef QtAQ
    }
    
    converged=0;
    
    /* Do QR iterations on the tridiagonalised matrix A.
     */
    #pragma acc loop seq
    for(int i=0; !converged && i<EIG_QR_MAX_ITS; i++){
        DEFINE_VARMATRIX_SYM3(rmat);
        DEFINE_VARMATRIX_3(qmat);
        #define Q(I) VAR_VARMATRIX(qmat, I)
        #define R(I) VAR_VARMATRIX(rmat, I)
        
        const int xy_small=fabs(A(I_XY)) <= fn_sqrt(fabs(A(I_XX))*fabs(A(I_YY)))*FNUMBER_EPS;
        const int yz_small=fabs(A(I_YZ)) <= fn_sqrt(fabs(A(I_YY))*fabs(A(I_ZZ)))*FNUMBER_EPS;
        if(xy_small || yz_small){
            if(!xy_small){
                /* Handle the 2X2 case
                 *
                 *        **0
                 * A_i =~ **0
                 *        00*
                 */
                FNumber l1, l2, v1, v2;
                DO_EIG2_LOWLEVEL(l1, l2, v1, v2, A(I_XX), A(I_YY), A(I_XY));
                
                A(I_XX)=l1;    A(I_YY)=l2; 
                Q(INS_XX)=v1;  Q(INS_YX)=v2;  Q(INS_ZX)=0;
                Q(INS_XY)=-v2; Q(INS_YY)=v1;  Q(INS_ZY)=0;
                Q(INS_XZ)=0;   Q(INS_YZ)=0;   Q(INS_ZZ)=1;
            }else if(!yz_small){
                /* Handle the 2X2 case
                 *
                 *        *00
                 * A_i =~ 0**
                 *        0**
                 */
                FNumber l1, l2, v1, v2;
                DO_EIG2_LOWLEVEL(l1, l2, v1, v2, A(I_YY), A(I_ZZ), A(I_YZ));

                A(I_YY)=l1;    A(I_ZZ)=l2;
                Q(INS_XX)=1;   Q(INS_YX)=0;   Q(INS_ZX)=0;
                Q(INS_XY)=0;   Q(INS_YY)=v1;  Q(INS_ZY)=v2;
                Q(INS_XZ)=0;   Q(INS_YZ)=-v2; Q(INS_ZZ)=v1;
            }
            A(I_XY)=0;     A(I_YZ)=0;
            /* We must wait until check in for statement to brea out of the loop:
             * the ev matrix update below is still needed.
             */
            converged=1;
        }else{
            /* Shift A by an eigenvalue of trailing 2x2 block 
             * to accelerate convergence
             */
            const FNumber w=A(I_YY), v=A(I_ZZ), u=A(I_YZ), d=(w-v)/2;
            const FNumber shift=v+d-sign(d)*fn_sqrt(d*d+u*u);
            
            A(I_XX)=A(I_XX)-shift;
            A(I_YY)=A(I_YY)-shift;
            A(I_ZZ)=A(I_ZZ)-shift;
            
            /* Calculate by Givens transformations A=QR, where R is upper
             * triangular and Q orthonormal.
             */
            {
                FNumber c1, s1, c2, s2;

                /* first two rows */
                GIVENS(c1, s1, A(I_XX), A(I_YX));
                R(I_XX)=c1*A(I_XX)-s1*A(I_YX);
                R(I_XY)=c1*A(I_XY)-s1*A(I_YY);
                R(I_XZ)=c1*A(I_XZ)-s1*A(I_YZ);
                const FNumber ryy=s1*A(I_XY)+c1*A(I_YY);
                const FNumber ryz=s1*A(I_XZ)+c1*A(I_YZ);
                
                /* last two rows */
                GIVENS(c2, s2, ryy, A(I_ZY));
                R(I_YY)=c2*ryy-s2*A(I_ZY);
                R(I_YZ)=c2*ryz-s2*A(I_ZZ);
                R(I_ZZ)=s2*ryz+c2*A(I_ZZ);

                /* Construct the orthonormal transformation matrix Q */
                Q(INS_XX)=c1;
                Q(INS_XY)=c2*s1;
                Q(INS_XZ)=s1*s2;
                Q(INS_YX)=-s1;
                Q(INS_YY)=c1*c2;
                Q(INS_YZ)=c1*s2;
                Q(INS_ZX)=0;
                Q(INS_ZY)=-s2;
                Q(INS_ZZ)=c2;
            }

            /* Update A(i+1) := R Q, where A(i)=Q R.
             * Note that R is upper triangular 
             *   ==> some multiplications unneceessary.
             * Note that A(i+1) remains symmetric 
             *   ==> some updates unnecessary
             *   ==> use the least expensive.
             */
            {
                A(I_XX)=R(I_XX)*Q(INS_XX)+R(I_XY)*Q(INS_YX)+R(I_XZ)*Q(INS_ZX);
                A(I_XY)=R(I_XX)*Q(INS_XY)+R(I_XY)*Q(INS_YY)+R(I_XZ)*Q(INS_ZY);
                /*A(I_XZ)=R(I_XX)*Q(INS_XZ)+R(I_XY)*Q(INS_YZ)+R(I_XZ)*Q(INS_ZZ);*/
                A(I_YY)=R(I_YY)*Q(INS_YY)+R(I_YZ)*Q(INS_ZY);
                /*A(I_YZ)=R(I_YY)*Q(INS_YZ)+R(I_YZ)*Q(INS_ZZ);*/
                A(I_ZY)=R(I_ZZ)*Q(INS_ZY);
                A(I_ZZ)=R(I_ZZ)*Q(INS_ZZ);
            }

            /* Shift A back.
             */
            A(I_XX)=A(I_XX)+shift;
            A(I_YY)=A(I_YY)+shift;
            A(I_ZZ)=A(I_ZZ)+shift;
        }
        
        /* If eigenvectors are to be calculated, update EV := EV Q in-place 
        */
        if(calc_evs){
            FNumber t1, t2, t3;
            
            t1=EV(INS_XX)*Q(INS_XX)+EV(INS_XY)*Q(INS_YX)+EV(INS_XZ)*Q(INS_ZX);
            t2=EV(INS_XX)*Q(INS_XY)+EV(INS_XY)*Q(INS_YY)+EV(INS_XZ)*Q(INS_ZY);
            t3=EV(INS_XX)*Q(INS_XZ)+EV(INS_XY)*Q(INS_YZ)+EV(INS_XZ)*Q(INS_ZZ);
            EV(INS_XX)=t1;
            EV(INS_XY)=t2;
            EV(INS_XZ)=t3;
            
            t1=EV(INS_YX)*Q(INS_XX)+EV(INS_YY)*Q(INS_YX)+EV(INS_YZ)*Q(INS_ZX);
            t2=EV(INS_YX)*Q(INS_XY)+EV(INS_YY)*Q(INS_YY)+EV(INS_YZ)*Q(INS_ZY);
            t3=EV(INS_YX)*Q(INS_XZ)+EV(INS_YY)*Q(INS_YZ)+EV(INS_YZ)*Q(INS_ZZ);
            EV(INS_YX)=t1;
            EV(INS_YY)=t2;
            EV(INS_YZ)=t3;
            
            t1=EV(INS_ZX)*Q(INS_XX)+EV(INS_ZY)*Q(INS_YX)+EV(INS_ZZ)*Q(INS_ZX);
            t2=EV(INS_ZX)*Q(INS_XY)+EV(INS_ZY)*Q(INS_YY)+EV(INS_ZZ)*Q(INS_ZY);
            t3=EV(INS_ZX)*Q(INS_XZ)+EV(INS_ZY)*Q(INS_YZ)+EV(INS_ZZ)*Q(INS_ZZ);
            EV(INS_ZX)=t1;
            EV(INS_ZY)=t2;
            EV(INS_ZZ)=t3;
        }
        
        #undef Q
        #undef R
    }

    /* Scale the eigenvalues back.
     */
    if(iscale){
        /*vecscale(amat, I_N, 1/sigma); */
        const FNumber rsigma=1/sigma;
        A(I_XX)*=rsigma;
        A(I_XY)*=rsigma;
        A(I_YY)*=rsigma;
        A(I_XZ)*=rsigma;
        A(I_YZ)*=rsigma;
        A(I_ZZ)*=rsigma;
    }
}

