/*
 * dticode/src/linoper.h
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */


#ifndef _LINOPER_H
#define _LINOPER_H

#include <libtu/types.h>
#include <libtu/obj.h>
#include "tensor.h"
#include "vector.h"
#include "number.h"
#include "diffoper.h"

INTRCLASS(LinOper);

DECLCLASS(LinOper){
    DOper doper;
};


typedef Vector LinOperDom;
typedef Vector LinOperCod;

DYNFUN void linoper_calculate(LinOper *op,
                              LinOperCod *y, 
                              const LinOperDom *x,
                              bool addto_y);

DYNFUN void linoper_calculatec(LinOper *op,
                               LinOperDom *y, 
                               const LinOperCod *x,
                               bool addto_y);

DYNFUN void linoper_backproj(LinOper *op,
                             LinOperDom *x, 
                             const LinOperCod *y);

DYNFUN void linoper_projdom(LinOper *op, LinOperDom *x);

// Calculate |f-Ax| and constraint m.
DYNFUN FNumber linoper_calcdist(LinOper *op,
                                const LinOperDom *x,
                                const LinOperCod *f,
                                FNumber *m);

DYNFUN LinOperDom *linoper_alloc_dom(LinOper *op);
DYNFUN LinOperCod *linoper_alloc_cod(LinOper *op);

DYNFUN FNumber linoper_estimate_norm(const LinOper *op);
DYNFUN FNumber linoper_estimate_lower_bound(const LinOper *op);

extern void linoper_deinit(LinOper *op);
extern bool linoper_init(LinOper *op);

// FidL2 support

INTRCLASS(LinOperL2);

DECLCLASS(LinOperL2){
    LinOper linoper;
};

// Solve for y: zeta y - xweight x + tau K^*(Ky-f)  \in -N_{constraint set}
extern void linoper_l2_gsolve(LinOperL2 *op,
                              LinOperDom *x,
                              FNumber tau, FNumber zeta, FNumber xweight,
                              const LinOperCod *f);

// Calculate |f-Ax|^2/2 and constraint m.
extern FNumber linoper_l2_g(LinOperL2 *op,
                            const LinOperDom *x,
                            const LinOperCod *f,
                            FNumber *m);

// Calculate G^*(x), with weight of constrained part split into c.
// x may be destroyed.
extern FNumber linoper_l2_gc(LinOperL2 *op,
                             const LinOperDom *minus_x,
                             const LinOperCod *f,
                             LinOperDom *xwork,
                             FNumber *c);

extern void linoper_l2_deinit(LinOperL2 *op);
extern bool linoper_l2_init(LinOperL2 *op);

#endif /* _LINOPER_H */
