/*
 * dticode/src/linoper_compose.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <assert.h>
#include <libtu/objp.h>
#include <libtu/types.h>
#include <libtu/misc.h>
#include <libtu/output.h>
#include "linoper.h"
#include "diffoper.h"
#include "linoper_compose.h"

void linoper_compose_calculate(LinOperCompose *op,
                               LinOperCod *y, 
                               const LinOperDom *x, 
                               bool addto_y)
{
    linoper_calculate(op->dop, op->tmp, x, FALSE);
    linoper_calculate(op->lop, y, op->tmp, addto_y);
}

void linoper_compose_backproj(LinOperCompose *op,
                               LinOperDom *x, 
                               const LinOperCod *y)
{
    linoper_backproj(op->lop, op->tmp, y);
    linoper_backproj(op->dop, x, op->tmp);
}

void linoper_compose_calculatec(LinOperCompose *op,
                                  LinOperDom *x, 
                                  const LinOperCod *y, 
                                  bool addto_x)
{
    linoper_calculatec(op->lop, op->tmp, y, FALSE);
    linoper_calculatec(op->dop, x, op->tmp, addto_x);
}


LinOperDom *linoper_compose_alloc_dom(LinOperCompose *op)
{
    return linoper_alloc_dom(op->dop);
}

LinOperCod *linoper_compose_alloc_cod(LinOperCompose *op)
{
    return linoper_alloc_cod(op->lop);
}

FNumber linoper_compose_constraint_scale(LinOperCompose *op, const LinOperDom *x)
{
    return diffoper_constraint_scale((DOper*)op->dop, x);
}

FNumber linoper_compose_constraint_conjugate(LinOperCompose *op, LinOperDom *x)
{
    return diffoper_constraint_conjugate((DOper*)op->dop, x);
}

void linoper_compose_projdom(LinOperCompose *op, LinOperDom *x)
{
    linoper_projdom(op->dop, x);
}

FNumber linoper_compose_estimate_norm(LinOperCompose *op)
{
    FNumber lopn, dopn;
    
    dopn=linoper_estimate_norm(op->dop);
    lopn=linoper_estimate_norm(op->lop);
        
    return lopn*dopn;
}

FNumber linoper_compose_estimate_lower_bound(LinOperCompose *op)
{
    FNumber lopn, dopn;
    
    dopn=linoper_estimate_lower_bound(op->dop);
    lopn=linoper_estimate_lower_bound(op->lop);
    
    return lopn*dopn;
}

bool linoper_compose_init(LinOperCompose *op, 
                          LinOper *lop, LinOper *dop)
{
    op->lop=lop;
    op->dop=dop;
    op->tmp=linoper_alloc_cod(dop);
    
    return linoper_init(&op->loper);
}

void linoper_compose_deinit(LinOperCompose *op)
{
    destroy_obj((Obj*)op->tmp);
    linoper_deinit(&op->loper);
}

LinOperCompose *create_linoper_compose(LinOper *lop, LinOper *dop)
{
    CREATEOBJ_IMPL(LinOperCompose, linoper_compose, (p, lop, dop));
}

static DynFunTab linoper_compose_dynfuntab[]={
    {linoper_calculate, linoper_compose_calculate},
    {linoper_backproj, linoper_compose_backproj},
    {linoper_calculatec, linoper_compose_calculatec},
    {linoper_projdom, linoper_compose_projdom},
    {(DynFun*)linoper_alloc_dom, (DynFun*)linoper_compose_alloc_dom},
    {(DynFun*)linoper_alloc_cod, (DynFun*)linoper_compose_alloc_cod},
    {(DynFun*)diffoper_constraint_scale, (DynFun*)linoper_compose_constraint_scale},
    {(DynFun*)diffoper_constraint_conjugate, (DynFun*)linoper_compose_constraint_conjugate},
    {(DynFun*)linoper_estimate_norm, (DynFun*)linoper_compose_estimate_norm},
    {(DynFun*)linoper_estimate_lower_bound, (DynFun*)linoper_compose_estimate_lower_bound},
    END_DYNFUNTAB,
};

IMPLCLASS(LinOperCompose, LinOper, linoper_compose_deinit, linoper_compose_dynfuntab);
