/*
 * dticode/src/linoper_grad.c
 *
 * Written 2014 by Tuomo Valkonen <tuomov@iki.fi>, EPN Quito.
 *
 */


#include <libtu/types.h>
#include <libtu/objp.h>
#include <libtu/misc.h>
#include <libtu/output.h>
#include "tensor.h"
#include "vector.h"
#include "number.h"
#include "linoper.h"
#include "linoper_grad.h"
#include "scaleplus.h"
#include "diff.h"
#include "tensornorm.h"
#include "sumdot1.h"

void linoper_grad_calculate(LinOperGrad *op,
                            LinOperGradCod *a, 
                            const LinOperGradDom *b,
                            bool addto_a)
{
    if(op->symmetric)
        tensorfield_symgrad(a, b, (addto_a ? 1.0 : 0.0), 1, 0, op->diffalg);
    else
        tensorfield_grad(a, b, (addto_a ? 1.0 : 0.0), 1, 0, op->diffalg);
}                                 

void linoper_grad_calculatec(LinOperGrad *op,
                             LinOperGradDom *a, 
                             const LinOperGradCod *b,
                             bool addto_a)
{
    if(op->symmetric)
        tensorfield_negsymdiv(a, b, (addto_a ? 1.0 : 0.0), 1, 0, op->diffalg);
    else
        tensorfield_negdiv(a, b, (addto_a ? 1.0 : 0.0), 1, 0, op->diffalg);
}

LinOperGradDom *linoper_grad_alloc_dom(LinOperGrad *op)
{
    return alloc_tensorfield(&op->sig);
}

LinOperGradCod *linoper_grad_alloc_cod(LinOperGrad *op)
{
    Sig dsig=op->sig; 
    dsig.order++; 
    dsig.symmetric=op->symmetric;
    
    return alloc_tensorfield(&dsig);
}

FNumber linoper_grad_estimate_norm(const LinOperGrad *op)
{
    return calc_l2fact(&(op->sig), op->diffalg);;
}

FNumber linoper_grad_estimate_lower_bound(const LinOperGrad *op)
{
    return 0;
}

void linoper_grad_deinit(LinOperGrad *op)
{
    linoper_deinit(&op->linoper);
}

bool linoper_grad_init(LinOperGrad *op, const Sig *sig, bool symmetric, const char *diffalg)
{
    linoper_init(&op->linoper);
    
    op->sig=*sig;
    op->symmetric=symmetric;
    op->diffalg=(diffalg[0] ? diffalg[0] : 'f');

    return TRUE;
}

LinOperGrad *alloc_linoper_grad(const Sig *sig, bool symmetric, const char *diffalg)
{
    CREATEOBJ_IMPL(LinOperGrad, linoper_grad, (p, sig, symmetric, diffalg));
}

static DynFunTab linoper_grad_dynfuntab[]={
    {linoper_calculate, linoper_grad_calculate},
    {linoper_calculatec, linoper_grad_calculatec},
    {(DynFun*)linoper_alloc_dom, (DynFun*)linoper_grad_alloc_dom},
    {(DynFun*)linoper_alloc_cod, (DynFun*)linoper_grad_alloc_cod},
    {(DynFun*)linoper_estimate_norm, (DynFun*)linoper_grad_estimate_norm},
    {(DynFun*)linoper_estimate_lower_bound, (DynFun*)linoper_grad_estimate_lower_bound},
    END_DYNFUNTAB,
};

IMPLCLASS(LinOperGrad, LinOper, linoper_grad_deinit, linoper_grad_dynfuntab);
