/*
 * dticode/src/linoper_id.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */


#include <libtu/types.h>
#include <libtu/objp.h>
#include <libtu/misc.h>
#include <libtu/output.h>
#include "tensor.h"
#include "vector.h"
#include "number.h"
#include "linoper.h"
#include "linoper_id.h"
#include "scaleplus.h"
#include "project_posdef.h"
#include "tensornorm.h"
#include "sumdot1.h"

void linoper_id_calculate(LinOperId *op,
                          LinOperIdCod *a, 
                          const LinOperIdDom *b,
                          bool addto_a)
{
    if(addto_a)
        tensorfield_scaleplus(a, a, b, 1, 1, 1);
    else
        tensorfield_copy(a, b);
}                                 

void linoper_id_calculatec(LinOperId *op,
                           LinOperIdDom *a, 
                           const LinOperIdCod *b,
                           bool addto_a)
{
    if(addto_a)
        tensorfield_scaleplus(a, a, b, 1, 1, 1);
    else
        tensorfield_copy(a, b);

}

void linoper_id_backproj(LinOperId *op,
                         LinOperIdDom *x, 
                         const LinOperIdCod *y)
{
    tensorfield_copy(x, y);
}                                

LinOperIdDom *linoper_id_alloc_dom(LinOperId *op)
{
    return alloc_tensorfield(&op->sig);
}

LinOperIdCod *linoper_id_alloc_cod(LinOperId *op)
{
    return alloc_tensorfield(&op->sig);
}

FNumber linoper_id_estimate_norm(const LinOperId *op)
{
    return 1;
}

FNumber linoper_id_estimate_lower_bound(const LinOperId *op)
{
    return 1;
}

void linoper_id_gsolve(LinOperId *op,
                       LinOperIdDom *u,
                       FNumber tau, FNumber zeta, FNumber cweight,
                       const LinOperIdCod *f)
{
    tensorfield_scaleplus(u, u, f, cweight, tau, zeta+tau);
}

FNumber linoper_id_g(LinOperId *op,
                     const LinOperIdDom *x,
                     const LinOperIdCod *f,
                     FNumber *m)
{
    if(m!=NULL)
        *m=0;
        
    return tensorfield_frob2dist22(f, x);
}                            
                              
// Calculate G^*(x), with weight of constrained part split into c.
FNumber linoper_id_gc(LinOperId *op,
                      const LinOperIdDom *minus_x,
                      const LinOperIdCod *f,
                      LinOperIdDom *xwork,
                      FNumber *c)
{
    if(c!=NULL)
        *c=0;
        
    tensorfield_scaleplus(xwork, f, minus_x, 1, -1, 1);
    return (-tensorfield_sumdot1(minus_x, xwork, FALSE, 1)
            -tensorfield_frob2dist22(f, xwork));
    
}

void linoper_id_deinit(LinOperId *op)
{
    linoper_l2_deinit(&op->linoperl2);
}

bool linoper_id_init(LinOperId *op, const Sig *sig)
{
    linoper_l2_init(&op->linoperl2);
    
    op->sig=*sig;
    
    return TRUE;
}

LinOperId *alloc_linoper_id(const Sig *sig)
{
    CREATEOBJ_IMPL(LinOperId, linoper_id, (p, sig));
}

static DynFunTab linoper_id_dynfuntab[]={
    {linoper_calculate, linoper_id_calculate},
    {linoper_calculatec, linoper_id_calculatec},
    {linoper_backproj, linoper_id_backproj},
    {linoper_l2_gsolve, linoper_id_gsolve},
    {(DynFun*)linoper_l2_g, (DynFun*)linoper_id_g},
    {(DynFun*)linoper_l2_gc, (DynFun*)linoper_id_gc},
    {(DynFun*)linoper_alloc_dom, (DynFun*)linoper_id_alloc_dom},
    {(DynFun*)linoper_alloc_cod, (DynFun*)linoper_id_alloc_cod},
    {(DynFun*)linoper_estimate_norm, (DynFun*)linoper_id_estimate_norm},
    {(DynFun*)linoper_estimate_lower_bound, (DynFun*)linoper_id_estimate_lower_bound},
    END_DYNFUNTAB,
};

IMPLCLASS(LinOperId, LinOperL2, linoper_id_deinit, linoper_id_dynfuntab);
