/*
 * dticode/src/linoper_multiply_by.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <assert.h>
#include <libtu/objp.h>
#include <libtu/types.h>
#include <libtu/misc.h>
#include <libtu/output.h>
#include "linoper_multiply_by.h"
#include "multiply.h"
#include "complexx.h"
#include "splitvar.h"
#include "tensornorm.h"
#include "normalise.h"


void linoper_multiply_by_calculate(LinOperMultiplyBy *op, 
                                   LinOperMultiplyByCod *y, 
                                   const LinOperMultiplyByDom *x,
                                   bool addto_y)
{
    tensorfield_multiply_pointwise(y, op->w, x, addto_y);
}

void linoper_multiply_by_backproj(LinOperMultiplyBy *op,
                                  LinOperMultiplyByDom *x, 
                                  const LinOperMultiplyByCod *y)
{
    tensorfield_divide_pointwise(x, y, op->w, FALSE);
}

void linoper_multiply_by_calculatec(LinOperMultiplyBy *op,
                                    LinOperMultiplyByDom *x, 
                                    const LinOperMultiplyByCod *y, 
                                    bool addto_x)
{
    tensorfield_multiply_pointwise_conj(x, op->w, y, addto_x);
}

LinOperMultiplyByDom *linoper_multiply_by_alloc_dom(LinOperMultiplyBy *op)
{
    return alloc_tensorfield(&op->sig);
}

LinOperMultiplyByCod *linoper_multiply_by_alloc_cod(LinOperMultiplyBy *op)
{
    Sig sig=op->sig;
    sig.is_complex=(sig.is_complex || op->w->sig.is_complex);
    
    return alloc_tensorfield(&sig);
}

FNumber linoper_multiply_by_estimate_norm(const LinOperMultiplyBy *op)
{
    return tensorfield_frobinf(op->w);
}

FNumber linoper_multiply_by_estimate_lower_bound(const LinOperMultiplyBy *op)
{
    #warning "TODO: better estimate"
    return 0;
}

bool linoper_multiply_by_init(LinOperMultiplyBy *op, 
                              const Sig *sig,
                              TensorField *w)
{
    assert(sig_tensor_count(&w->sig)==sig_tensor_count(sig));
    assert(sig_tensor_size_c(&w->sig)==1 || sig_tensor_size(&w->sig)==sig_tensor_size(sig));
    
    op->sig=*sig;
    op->w=w;
    
    return linoper_init(&op->loper);
}

void linoper_multiply_by_deinit(LinOperMultiplyBy *op)
{
    destroy_obj((Obj*)op->w);
    linoper_deinit(&op->loper);
}

LinOperMultiplyBy *create_linoper_multiply_by(const Sig *sig,
                                              TensorField *w)
{
    CREATEOBJ_IMPL(LinOperMultiplyBy, linoper_multiply_by, (p, sig, w));
}

static DynFunTab linoper_multiply_by_dynfuntab[]={
    {linoper_calculate, linoper_multiply_by_calculate},
    {linoper_backproj, linoper_multiply_by_backproj},
    {linoper_calculatec, linoper_multiply_by_calculatec},
    {(DynFun*)linoper_alloc_dom, (DynFun*)linoper_multiply_by_alloc_dom},
    {(DynFun*)linoper_alloc_cod, (DynFun*)linoper_multiply_by_alloc_cod},
    {(DynFun*)linoper_estimate_norm, (DynFun*)linoper_multiply_by_estimate_norm},
    {(DynFun*)linoper_estimate_lower_bound, (DynFun*)linoper_multiply_by_estimate_lower_bound},
    END_DYNFUNTAB,
};

IMPLCLASS(LinOperMultiplyBy, LinOper, linoper_multiply_by_deinit, linoper_multiply_by_dynfuntab);
