/*
 * dticode/src/linoper_nufft.h
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */


#ifndef _LINOPER_NUFFT_H
#define _LINOPER_NUFFT_H

#include <libtu/types.h>
#include <libtu/obj.h>
#include <nfft3.h>
#include "tensor.h"
#include "vector.h"
#include "number.h"
#include "linoper.h"

#define NUFFT_IDX(op) (op->idx)
#define NUFFT_N(op) (op->n_samples)

typedef TensorField LinOperNUFFTCod;
typedef TensorField LinOperNUFFTDom;

INTRCLASS(LinOperNUFFT);

DECLCLASS(LinOperNUFFT){
    LinOper linoper;
    TensorField *f;
    size_t n_samples;
    nfft_plan plan;
    Sig sig;
    LinOperNUFFTCod *tmp_cod;
    LinOperNUFFTDom *tmp_dom;
    bool is_copy;
};

extern void linoper_nufft_calculate(LinOperNUFFT *op,
                                    LinOperNUFFTCod *y, 
                                    const LinOperNUFFTDom *x,
                                    bool addto_y);

extern void linoper_nufft_calculatec(LinOperNUFFT *op,
                                     LinOperNUFFTDom *y, 
                                     const LinOperNUFFTCod *x,
                                     bool addto_y);

extern void linoper_nufft_backproj(LinOperNUFFT *op,
                                   LinOperNUFFTDom *x, 
                                   const LinOperNUFFTCod *y);

extern LinOperNUFFTDom *linoper_nufft_alloc_dom(LinOperNUFFT *op);
extern LinOperNUFFTCod *linoper_nufft_alloc_cod(LinOperNUFFT *op);

extern FNumber linoper_nufft_estimate_norm(const LinOperNUFFT *op);

extern void linoper_nufft_gsolve(LinOperNUFFT *op,
                                 LinOperNUFFTDom *x,
                                 FNumber tau, FNumber zeta, FNumber xweight,
                                 const LinOperNUFFTCod *f);

extern void linoper_nufft_deinit(LinOperNUFFT *op);

extern bool linoper_nufft_init(LinOperNUFFT *op, 
                               const Sig *sig, 
                               size_t n_samples);

extern LinOperNUFFT *alloc_linoper_nufft(const Sig *sig, 
                                         size_t n_samples);

extern LinOperNUFFT *load_linoper_nufft(const char *infile, 
                                        const Sig *sig, 
                                        TensorField **f_);

extern LinOperNUFFT *load_linoper_nufft_nodata(const char *infile, 
                                               const Sig *sig, 
                                               Sig *fsig);

extern LinOperNUFFT *linoper_nufft_copy(const LinOperNUFFT *op);

#endif /* _LINOPER_NUFFT_H */
