/*
 * dticode/src/linoper_pointwise.h
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */


#ifndef _LINOPER_POINTWISE_H
#define _LINOPER_POINTWISE_H

#include <libtu/types.h>
#include <libtu/obj.h>
#include "tensor.h"
#include "vector.h"
#include "number.h"
#include "linoper.h"
#include "matrix.h"

typedef TensorField LinOperPointwiseCod;
typedef TensorField LinOperPointwiseDom;

INTRCLASS(LinOperPointwise);
INTRCLASS(LinOperPointwiseNonNeg);

DECLCLASS(LinOperPointwise){
    LinOperL2 linoperl2;
    const Matrix *amat;
    Matrix *amat_conj;
    Sig sig;
    FNumber iprho;
    FNumber a_squared_bound;
};

DECLCLASS(LinOperPointwiseNonNeg){
    LinOperPointwise linoperpw;
};

extern void linoper_pointwise_calculate(LinOperPointwise *op,
                                        LinOperPointwiseCod *y, 
                                        const LinOperPointwiseDom *x,
                                        bool addto_y);

extern void linoper_pointwise_calculatec(LinOperPointwise *op,
                                         LinOperPointwiseDom *y, 
                                         const LinOperPointwiseCod *x,
                                         bool addto_y);

extern void linoper_pointwise_backproj(LinOperPointwise *op,
                                       LinOperPointwiseDom *x, 
                                       const LinOperPointwiseCod *y);

extern void linoper_pointwise_gsolve(LinOperPointwise *op,
                                     LinOperPointwiseDom *x,
                                     FNumber tau, FNumber zeta, FNumber cweight,
                                     const LinOperPointwiseCod *f);

extern LinOperPointwiseDom *linoper_pointwise_alloc_dom(LinOperPointwise *op);
extern LinOperPointwiseCod *linoper_pointwise_alloc_cod(LinOperPointwise *op);

extern FNumber linoper_pointwise_estimate_norm(const LinOperPointwise *op);

extern void linoper_pointwise_deinit(LinOperPointwise *op);

extern bool linoper_pointwise_init(LinOperPointwise *op, 
                                   const Sig *sig,
                                   const Matrix *amat);

extern LinOperPointwise *alloc_linoper_pointwise(const Sig *sig, 
                                                 const Matrix *amat);

extern void linoper_pointwise_nonneg_deinit(LinOperPointwiseNonNeg *op);

extern bool linoper_pointwise_nonneg_init(LinOperPointwiseNonNeg *op, 
                                          const Sig *sig,
                                          const Matrix *amat);

extern LinOperPointwiseNonNeg *alloc_linoper_pointwise_nonneg(const Sig *sig, 
                                                              const Matrix *amat);

#endif /* _LINOPER_POINTWISE_H */
