/*
 * dticode/src/linoper_sparsefft.h
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */


#ifndef _LINOPER_SPARSEFFT_H
#define _LINOPER_SPARSEFFT_H

#include <libtu/types.h>
#include <libtu/obj.h>
#include "tensor.h"
#include "vector.h"
#include "number.h"
#include "linoper.h"
#include "fft.h"

#define SPARSEFFT_IDX(op) (op->idx)
#define SPARSEFFT_N(op) (op->n_samples)

typedef TensorField LinOperSparseFFTCod;
typedef TensorField LinOperSparseFFTDom;

INTRCLASS(LinOperSparseFFT);

DECLCLASS(LinOperSparseFFT){
    LinOperL2 linoperl2;
    TensorField *f;
    size_t n_samples;
    size_t * DEVPTR idx;
    Sig sig;
    bool full;
    LinOperSparseFFTDom *tmp;
};

extern void linoper_sparsefft_calculate(LinOperSparseFFT *op,
                                        LinOperSparseFFTCod *y, 
                                        const LinOperSparseFFTDom *x,
                                        bool addto_y);

extern void linoper_sparsefft_calculatec(LinOperSparseFFT *op,
                                         LinOperSparseFFTDom *y, 
                                         const LinOperSparseFFTCod *x,
                                         bool addto_y);

extern void linoper_sparsefft_backproj(LinOperSparseFFT *op,
                                       LinOperSparseFFTDom *x, 
                                       const LinOperSparseFFTCod *y);

extern LinOperSparseFFTDom *linoper_sparsefft_alloc_dom(LinOperSparseFFT *op);
extern LinOperSparseFFTCod *linoper_sparsefft_alloc_cod(LinOperSparseFFT *op);

extern FNumber linoper_sparsefft_estimate_norm(const LinOperSparseFFT *op);

extern void linoper_sparsefft_gsolve(LinOperSparseFFT *op,
                                     LinOperSparseFFTDom *x,
                                     FNumber tau, FNumber zeta, FNumber xweight,
                                     const LinOperSparseFFTCod *f);

extern void linoper_sparsefft_deinit(LinOperSparseFFT *op);

extern bool linoper_sparsefft_init(LinOperSparseFFT *op, 
                                   const Sig *sig, 
                                   size_t n_samples);

extern LinOperSparseFFT *alloc_linoper_sparsefft(const Sig *sig, 
                                                 size_t n_samples);

extern LinOperSparseFFT *load_linoper_sparsefft(const char *infile, 
                                                const Sig *sig, 
                                                bool sparse_infile,
                                                TensorField **f_);

extern LinOperSparseFFT *load_linoper_sparsefft_nodata(const char *infile, 
                                                       const Sig *sig, 
                                                       bool sparse_infile,
                                                       Sig *fsig);

extern LinOperSparseFFT *linoper_sparsefft_copy(const LinOperSparseFFT *copyop);

#endif /* _LINOPER_SPARSEFFT_H */
