/*
 * dticode/src/linoper_sparsemat.h
 *
 * Written 2014 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */


#ifndef _LINOPER_SPARSEMAT_H
#define _LINOPER_SPARSEMAT_H

#include <libtu/types.h>
#include <libtu/obj.h>
#include "tensor.h"
#include "vector.h"
#include "number.h"
#include "linoper.h"
#include "sparse.h"

#define SPARSEMAT_IDX(op) (op->idx)
#define SPARSEMAT_N(op) (op->n_samples)

typedef TensorField LinOperSparseMatCod;
typedef TensorField LinOperSparseMatDom;

INTRCLASS(LinOperSparseMat);

DECLCLASS(LinOperSparseMat){
    LinOper linoper;
    SparseMatrix pm;
    Sig sigdom;
    Sig sigcod;
    FNumber lbnd;
    FNumber ubnd;
};

extern void linoper_sparsemat_calculate(LinOperSparseMat *op,
                                        LinOperSparseMatCod *y, 
                                        const LinOperSparseMatDom *x,
                                        bool addto_y);

extern void linoper_sparsemat_calculatec(LinOperSparseMat *op,
                                         LinOperSparseMatDom *y, 
                                         const LinOperSparseMatCod *x,
                                         bool addto_y);

extern void linoper_sparsemat_backproj(LinOperSparseMat *op,
                                       LinOperSparseMatDom *x, 
                                       const LinOperSparseMatCod *y);

extern LinOperSparseMatDom *linoper_sparsemat_alloc_dom(LinOperSparseMat *op);
extern LinOperSparseMatCod *linoper_sparsemat_alloc_cod(LinOperSparseMat *op);

extern FNumber linoper_sparsemat_estimate_norm(const LinOperSparseMat *op);

extern void linoper_sparsemat_gsolve(LinOperSparseMat *op,
                                     LinOperSparseMatDom *x,
                                     FNumber tau, FNumber zeta, FNumber xweight,
                                     const LinOperSparseMatCod *f);

extern void linoper_sparsemat_deinit(LinOperSparseMat *op);

extern bool linoper_sparsemat_init(LinOperSparseMat *op, 
                                   const Sig *sigcod, 
                                   const Sig *sigdom,
                                   SparseMatrix *pm,
                                   FNumber lbnd,
                                   FNumber ubnd);

extern LinOperSparseMat *alloc_linoper_sparsemat(const Sig *sigcod, 
                                                 const Sig *sigdom,
                                                 SparseMatrix *pm,
                                                 FNumber lbnd,
                                                 FNumber ubnd);

extern LinOperSparseMat *load_linoper_sparsemat(const char *infile, 
                                                const Sig *sigcod, 
                                                const Sig *sigdom);

#endif /* _LINOPER_SPARSEMAT_H */
