/*
 * dticode/src/lusolve_core.h
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */

// GETM/SETM(a, i, j): i=column, j=row !

{
    const FNumber smlnum=FNUMBER_SFMIN/FNUMBER_PREC;
    const FNumber bignum=FNUMBER(1.0)/smlnum;
    const FNumber rmin=fn_sqrt(smlnum);
    const FNumber rmax=fn_sqrt(bignum);

#if 0
    // Scale the matrix A to safe range.
    FNumber sigma=1; 
    int iscale=0;
    {
        FNumber anrm=0;
        for(size_t i=0; i<LUDIM; i++){
            for(size_t j=0; j<LUDIM; j++)
                anrm=fn_max(GETM(A, i, j), anrm);
        }
        
        if(anrm>0 && anrm<rmin){
            iscale=1;
            sigma=rmin/anrm;
        }else if(anrm>rmax){
            iscale=1;
            sigma=rmax/anrm;
        }
        for(size_t i=0; i<LUDIM; i++){
            for(size_t j=0; j<LUDIM; j++)
                SETM(A, i, j, sigma*GETM(A, i, j));
        }
    }
#endif

    // Loop over columns
    for(size_t i=0; i<LUDIM; i++){
        // Do pivoting
        FNumber mx=fn_abs(GETM(A, i, i));
        size_t mxi=i;
        for(size_t j=i+1; j<LUDIM; j++){
            FNumber b=fn_abs(GETM(A, i, j));
            if(b>mx){
                mx=b;
                mxi=j;
            }
        }
        if(mxi!=i){
            // Exchange rows mxi and i
            for(size_t k=i; k<LUDIM; k++){
                FNumber tmp=GETM(A, k, i);
                SETM(A, k, i, GETM(A, k, mxi));
                SETM(A, k, mxi, tmp);
            }
#if LUCORE_RHS_PARALLEL
            #pragma acc kernels loop deviceptr(C), independent
            #pragma omp parallel for
#else
            #pragma acc loop independent
#endif
            for(size_t rhs=0; rhs<NRHS; rhs++){
                FNumber tmp=GETV(C, i, rhs);
                SETV(C, i, rhs, GETV(C, mxi, rhs));
                SETV(C, mxi, rhs, tmp);
            }
        } 
        //printmat("Ak_pivot", LUDIM, a);
        
        // Loop over row > column
        for(size_t j=i+1; j<LUDIM; j++){
            FNumber eta=GETM(A, i, j)/GETM(A, i, i);
            if(eta!=0){
                // Update A := M A, where the transformation 
                // matrix M subtract eta-scaled i-row from j-row.
                // Leading columns < i should be zero here.
#if !LUCORE_RHS_PARALLEL
                #pragma acc loop independent
#endif
                for(size_t k=i; k<LUDIM; k++)
                    SETM(A, k, j, GETM(A, k, j)-eta*GETM(A, k, i));
                // Update right-hand-side c := M^(-1) c, 
#if LUCORE_RHS_PARALLEL
                #pragma acc kernels loop deviceptr(C), independent
                #pragma omp parallel for
#else
                #pragma acc loop independent
#endif
                for(size_t rhs=0; rhs<NRHS; rhs++){
                    SETV(C, j, rhs, GETV(C, j, rhs)-eta*GETV(C, i, rhs));
                }
            }
        }
        //printmat("Ak_result", LUDIM, a);
    }
    
    //printvec("cres", LUDIM, c);
    
    // Now A is upper triangular => Do Gaussian elimination
    
#if LUCORE_RHS_PARALLEL
    #pragma acc kernels loop deviceptr(C,RES), \
            copyin(A[0:LUDIM*LUDIM]), \
            independent
    #pragma omp parallel for
#else
    #pragma acc loop independent
#endif
    for(size_t rhs=0; rhs<NRHS; rhs++){
        // Loop over rows
        #pragma acc loop independent
        for(int j=LUDIM-1; j>=0; j--){
            FNumber tmp=0;
            // Loop over (the non-zero) columns > j in current row
            // to compute already known part of the left hand side.
            #pragma acc loop vector, reduction(+:tmp)
            for(size_t i=j+1; i<LUDIM; i++)
                tmp+=GETM(A, i, j)*GETV(RES, i, rhs);
            // Solve res[j]
            SETV(RES, j, rhs, (GETV(C, j, rhs)-tmp)/GETM(A, j, j));
        }
#if 0
        // Scale res if A was scaled
        if(iscale==1){
            const FNumber rsigma=1/sigma;
            for(size_t i=0; i<LUDIM; i++)
                SETV(RES, i, rhs, rsigma*GETV(RES, i, rhs));
        }
#endif
    }
    
}
