/*
 * dticode/src/matrix.h
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */


#ifndef _MATRIX_H
#define _MATRIX_H

#include "acc_support.h"
#include "tensor.h"
#include "number.h"

typedef struct{
    size_t n; // rows
    size_t m; // columns
    Sig sign, sigm;
    FNumber *p;
    bool is_identity;
} Matrix;

extern Matrix *alloc_matrix(const Sig *sign, 
                            const Sig *sigm);
extern Matrix *load_matrix(const Sig *sign, 
                           const Sig *sigm,
                           const char *file);
                           
extern Matrix *matrix_conjugate(const Matrix *m);
extern Matrix *matrix_multiply(const Matrix *a, const Matrix *b);
extern FNumber matrix_opnorm(const Matrix *a);
extern FNumber matrix_normsquared_lowerbound(const Matrix *a);

extern void tensorfield_linop(TensorField *y,
                              const Matrix *a,
                              const TensorField *x,
                              FNumber yfact,
                              FNumber xfact);

extern FNumber tensorfield_linop_frob2dist22(const TensorField *y,
                                             const Matrix *a,
                                             const TensorField *x);

extern void free_matrix(Matrix *m);


#endif /* _MATRIX_H */
