/*
 * dticode/src/matrix_misc.h
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */

#ifndef _MATRIX_MISC_H
#define _MATRIX_MISC_H

#include "number.h"

static void print_symm(const char *l, FNumber *a)
{
    printf("--\n");
    printf("%s1: %10.3g, %10.3g, %10.3g\n", l, a[I_XX], a[I_XY], a[I_XZ]);
    printf("%s2: %10.3g, %10.3g, %10.3g\n", l, a[I_YX], a[I_YY], a[I_YZ]);
    printf("%s3: %10.3g, %10.3g, %10.3g\n", l, a[I_ZX], a[I_ZY], a[I_ZZ]);
}

static void print_uppertri(const char *l, FNumber *a)
{
    printf("--\n");
    printf("%s1: %10.3g, %10.3g, %10.3g\n", l, a[I_XX], a[I_XY], a[I_XZ]);
    printf("%s2: %10.3g, %10.3g, %10.3g\n", l, 0.0,     a[I_YY], a[I_YZ]);
    printf("%s3: %10.3g, %10.3g, %10.3g\n", l, 0.0,     0.0,     a[I_ZZ]);
}

static void print_nonsymm(const char *l, FNumber *q)
{
    printf("--\n");
    printf("%s1: %10.3g, %10.3g, %10.3g\n", l, q[INS_XX], q[INS_XY], q[INS_XZ]);
    printf("%s2: %10.3g, %10.3g, %10.3g\n", l, q[INS_YX], q[INS_YY], q[INS_YZ]);
    printf("%s3: %10.3g, %10.3g, %10.3g\n", l, q[INS_ZX], q[INS_ZY], q[INS_ZZ]);
}

static void print_sparse(char *name, const SparseMatrix *sp)
{
    size_t i, j, end2;
    fprintf(stderr, "%s\n", name);
    for(i=0; i<sp->n; i++){
        end2=sp->jc[i+1];
        for(j=sp->jc[i]; j<end2; j++){
            fprintf(stderr, "(%d, %d) %e\n", (int) sp->ir[j]+1, (int) i+1, sp->pr[j]);
        }
    }
}

#define SWAP(a, I, J, tmp) tmp=a[I]; a[I]=a[J]; a[J]=tmp;

static void transpose(FNumber *a)
{
    FNumber tmp;
    
    SWAP(a, INS_XY, INS_YX, tmp);
    SWAP(a, INS_XZ, INS_ZX, tmp);
    SWAP(a, INS_YZ, INS_ZY, tmp);
}

static void multiply_nonsymm_nonsymm(FNumber *a, FNumber *r, FNumber *q)
{
    a[INS_XX]=r[INS_XX]*q[INS_XX]+r[INS_XY]*q[INS_YX]+r[INS_XZ]*q[INS_ZX];
    a[INS_XY]=r[INS_XX]*q[INS_XY]+r[INS_XY]*q[INS_YY]+r[INS_XZ]*q[INS_ZY];
    a[INS_XZ]=r[INS_XX]*q[INS_XZ]+r[INS_XY]*q[INS_YZ]+r[INS_XZ]*q[INS_ZZ];
    a[INS_YX]=r[INS_YX]*q[INS_XX]+r[INS_YY]*q[INS_YX]+r[INS_YZ]*q[INS_ZX];
    a[INS_YY]=r[INS_YX]*q[INS_XY]+r[INS_YY]*q[INS_YY]+r[INS_YZ]*q[INS_ZY];
    a[INS_YZ]=r[INS_YX]*q[INS_XZ]+r[INS_YY]*q[INS_YZ]+r[INS_YZ]*q[INS_ZZ];
    a[INS_ZX]=r[INS_ZX]*q[INS_XX]+r[INS_ZY]*q[INS_YX]+r[INS_ZZ]*q[INS_ZX];
    a[INS_ZY]=r[INS_ZX]*q[INS_XY]+r[INS_ZY]*q[INS_YY]+r[INS_ZZ]*q[INS_ZY];
    a[INS_ZZ]=r[INS_ZX]*q[INS_XZ]+r[INS_ZY]*q[INS_YZ]+r[INS_ZZ]*q[INS_ZZ];
}

static void multiply_symm_nonsymm(FNumber *a, FNumber *r, FNumber *q)
{
    a[INS_XX]=r[I_XX]*q[INS_XX]+r[I_XY]*q[INS_YX]+r[I_XZ]*q[INS_ZX];
    a[INS_XY]=r[I_XX]*q[INS_XY]+r[I_XY]*q[INS_YY]+r[I_XZ]*q[INS_ZY];
    a[INS_XZ]=r[I_XX]*q[INS_XZ]+r[I_XY]*q[INS_YZ]+r[I_XZ]*q[INS_ZZ];
    a[INS_YX]=r[I_YX]*q[INS_XX]+r[I_YY]*q[INS_YX]+r[I_YZ]*q[INS_ZX];
    a[INS_YY]=r[I_YX]*q[INS_XY]+r[I_YY]*q[INS_YY]+r[I_YZ]*q[INS_ZY];
    a[INS_YZ]=r[I_YX]*q[INS_XZ]+r[I_YY]*q[INS_YZ]+r[I_YZ]*q[INS_ZZ];
    a[INS_ZX]=r[I_ZX]*q[INS_XX]+r[I_ZY]*q[INS_YX]+r[I_ZZ]*q[INS_ZX];
    a[INS_ZY]=r[I_ZX]*q[INS_XY]+r[I_ZY]*q[INS_YY]+r[I_ZZ]*q[INS_ZY];
    a[INS_ZZ]=r[I_ZX]*q[INS_XZ]+r[I_ZY]*q[INS_YZ]+r[I_ZZ]*q[INS_ZZ];
}

static void multiply_nonsymm_symm(FNumber *a, FNumber *q, FNumber *r)
{
    a[INS_XX]=q[INS_XX]*r[I_XX]+q[INS_XY]*r[I_YX]+q[INS_XZ]*r[I_ZX];
    a[INS_XY]=q[INS_XX]*r[I_XY]+q[INS_XY]*r[I_YY]+q[INS_XZ]*r[I_ZY];
    a[INS_XZ]=q[INS_XX]*r[I_XZ]+q[INS_XY]*r[I_YZ]+q[INS_XZ]*r[I_ZZ];
    a[INS_YX]=q[INS_YX]*r[I_XX]+q[INS_YY]*r[I_YX]+q[INS_YZ]*r[I_ZX];
    a[INS_YY]=q[INS_YX]*r[I_XY]+q[INS_YY]*r[I_YY]+q[INS_YZ]*r[I_ZY];
    a[INS_YZ]=q[INS_YX]*r[I_XZ]+q[INS_YY]*r[I_YZ]+q[INS_YZ]*r[I_ZZ];
    a[INS_ZX]=q[INS_ZX]*r[I_XX]+q[INS_ZY]*r[I_YX]+q[INS_ZZ]*r[I_ZX];
    a[INS_ZY]=q[INS_ZX]*r[I_XY]+q[INS_ZY]*r[I_YY]+q[INS_ZZ]*r[I_ZY];
    a[INS_ZZ]=q[INS_ZX]*r[I_XZ]+q[INS_ZY]*r[I_YZ]+q[INS_ZZ]*r[I_ZZ];
}

static void multiply_nonsymm_uppertri(FNumber *a, FNumber *q, FNumber *r)
{
    a[INS_XX]=q[INS_XX]*r[I_XX];
    a[INS_XY]=q[INS_XX]*r[I_XY]+q[INS_XY]*r[I_YY];
    a[INS_XZ]=q[INS_XX]*r[I_XZ]+q[INS_XY]*r[I_YZ]+q[INS_XZ]*r[I_ZZ];
    a[INS_YX]=q[INS_YX]*r[I_XX];
    a[INS_YY]=q[INS_YX]*r[I_XY]+q[INS_YY]*r[I_YY];
    a[INS_YZ]=q[INS_YX]*r[I_XZ]+q[INS_YY]*r[I_YZ]+q[INS_YZ]*r[I_ZZ];
    a[INS_ZX]=q[INS_ZX]*r[I_XX];
    a[INS_ZY]=q[INS_ZX]*r[I_XY]+q[INS_ZY]*r[I_YY];
    a[INS_ZZ]=q[INS_ZX]*r[I_XZ]+q[INS_ZY]*r[I_YZ]+q[INS_ZZ]*r[I_ZZ];
}

#endif /* _MATRIX_MISC_H */
