/*
 * dticode/src/method.h
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */


#ifndef _METHOD_H
#define _METHOD_H

#include <libtu/optparser.h>
#include "number.h"
#include "cpock_accel.h"
#include "combine.h"
#include "combine_nonlin.h"
#include "tensor.h"

typedef struct{
    CPockParam cpockp;
    Sig sig;
    bool gaussnewton;
    bool bregman;
    uint maxfbiter;
    FNumber disttarget;
    FNumber fbeps;
    const char *diffalg;
    FidInitPrimalStyle initstyle;
    const char *initfile;
    const char *outfile;
    bool timing;
    bool fullsave;
} MethodParam;

#define DEFAULT_CPOCKP { \
    .sigma = -1.0, /* => 0.95/norm(K)*/ \
    .tau = -1.0, /* => 0.95/norm(K) */ \
    .gamma = INFINITY, /* set by function choice */ \
    .rho = 0.001, \
    .theta = 1, /* Not used by algorithm if gamma > 0 */ \
    .knorm = 0, /* Set by algorithm choice */ \
    .maxiter = 10000, \
    .verbose = FALSE, \
    .verbose_granularity = 1, \
    .calculate_pdgap0 = TRUE, \
    .mode = NLCPockInterp, \
    }

#define DEFAULT_SIG { \
        .dim = 2, \
        .order = 0, \
        .symmetric = FALSE, \
        .is_complex = FALSE, \
        .w = 0, \
        .h = 0, \
        .d = 1, \
        .xres = 1, \
        .yres = 1, \
        .zres = 1 \
    }

#define DEFAULT_METHOD_PARAM { \
    .cpockp = DEFAULT_CPOCKP, \
    .sig = DEFAULT_SIG, \
    .gaussnewton = FALSE, \
    .bregman = FALSE, \
    .maxfbiter = 1, \
    .fbeps = 0, \
    .disttarget = 0, \
    .diffalg = "f", \
    .initstyle = FidInitPrimalZero, \
    .initfile = NULL, \
    .outfile = NULL, \
    .timing = FALSE, \
    .fullsave = FALSE, \
    }

#define QUICK_OPT(X, Y, Z) {OPT_ID(X), Y, OPT_ARG, Y, Z}

#define METHOD_OPTS \
    QUICK_OPT('w', "width",   "Data domain width"), \
    QUICK_OPT('h', "height",  "Data domain height"), \
    QUICK_OPT('d', "depth",   "Data domain depth (default: 1)"), \
    QUICK_OPT('X', "xres",    "Data X sample distance (default: 1)"), \
    QUICK_OPT('Y', "yres",    "Data Y sample distance (default: 1)"), \
    QUICK_OPT('Z', "zres",    "Data Z sample distance (default: 1)"), \
    QUICK_OPT('t', "tau",     "Chambolle-Pock tau parameter"), \
    QUICK_OPT('s', "sigma",   "Chambolle-Pock sigma parameter"), \
    QUICK_OPT('P', "nlcpmode","Nonlinear Chambolle-Pock mode: lin|exact|interp"), \
    QUICK_OPT('M', "method",  "Method: cpock|gn|bregman-cpock|bregman-gn"), \
    QUICK_OPT('r', "rho",     "Target fractional duality gap"), \
    QUICK_OPT('i', "maxiter", "Maximum Chambolle-Pock iterations"), \
    QUICK_OPT('g', "diffalg", "Differentation method: f|c|b"), \
    QUICK_OPT('I', "maxfbiter", "Max. feedback (Bregman, Gauss-Newton) iterations"), \
    QUICK_OPT('B', "disttarget", "Do feedback iterations until distance below parameter"), \
    QUICK_OPT('e', "fbeps", "Feedback iteration stopping accuracy"), \
    QUICK_OPT('o', "outfile", "Output file prefix (raw doubles)"), \
    {OPT_ID('p'), "fullsave", 0, NULL, "Save all (dual and extra primal) variables"}, \
    QUICK_OPT('4', "initstyle", "Init style: zero|backproj|file"), \
    QUICK_OPT('3', "initfile",  "Initialisation file prefix (raw doubles)"), \
    {OPT_ID('T'), "timing", 0, NULL, "Display timing information"}, \
    {OPT_ID('V'), "verbose", 0, NULL, "Display iteration information "}, \
    QUICK_OPT('G', "vg",      "Verbose granularity (default: 1)"), \
    {'h', "help", 0, NULL, "Show this help"}

bool method_do_opt(MethodParam *mp, int opt, OptParserOpt *opts);

void method_params_sanity_check(MethodParam *mp);

void method_do(MethodParam *mp, Fid *fid, FidParams *fidp, Reg *reg, RegParams *regp);
void method_do_nl(MethodParam *mp, FidNL *fid, FidParams *fidp, Reg *reg, RegParams *regp);

#endif /* _METHOD_H */
