/*
 * dticode/src/misc.h
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */


#ifndef _MISC_H
#define _MISC_H

#include <math.h>

#include "number.h"

/* */

static __inline__ FNumber powerdiff(FNumber a, FNumber b)
{
    FNumber d=a-b;
    return d*d;
}

static __inline__ FNumber mult(FNumber a, FNumber b)
{
    return a*b;
}

/* PGCC doesn't generate a reduction if we pass use an inline function
 * plus. We have to use a macro. fmax is likewise detected for reduction.
 */
#define plus(A, B) (A + B)
#define idop(A) A

#define UPDATE_PLUS 1
#define UPDATE_MAX 2

/* */

static __inline__ FNumber sign(FNumber a)
{
    if(a<0)
        return -1;
    else
        return 1;
}

/* */

#if !_OPENACC

static __inline__ void veczero(FNumber *p, size_t n)
{
    for(size_t i=0; i<n; i++)
        p[i]=0;
}

static __inline__ void veccopy(FNumber *p, const FNumber *q, size_t n)
{
    for(size_t i=0; i<n; i++)
        p[i]=q[i];
}

static __inline__ void vecscale(FNumber *a, FNumber beta, size_t n)
{
    for(size_t i=0; i<n; i++)
        a[i]=a[i]*beta;
}

static __inline__ void vecadd(FNumber *r, const FNumber *a, 
                              FNumber alpha, size_t sz)
{
    for(size_t i=0; i<sz; i++)
        r[i]+=alpha*a[i];
}

#else

#define veczero(p, n) \
{ \
    _Pragma("acc loop independent") \
    for(size_t i_=0; i_<n; i_++) \
        (p)[i_]=0; \
}

#define veccopy(p, q, n) \
{ \
    _Pragma("acc loop independent") \
    for(size_t i_=0; i_<n; i_++) \
        (p)[i_]=(q)[i_]; \
}

#define vecscale(a, beta, n) \
{ \
    _Pragma("acc loop independent") \
    for(size_t i_=0; i_<n; i_++) \
        (a)[i_]=(a)[i_]*beta; \
}

#define vecadd(r, a, alpha, sz) \
{ \
    _Pragma("acc loop independent") \
    for(size_t i_=0; i_<sz; i_++) \
        (r)[i_]+=alpha*(a)[i_]; \
}

#endif


/*
static __inline__ FNumber vecmaxabs(FNumber *a, size_t n)
{
    FNumber m=0;
    
    for(size_t i=0; i<n; i++)
        m=fmax(m, fabs(a[i]));
        
    return m;
}
*/

#define MATVECMULT(r, a, v, n, m) \
    { \
        _Pragma("acc loop independent") \
        for(size_t _i=0; _i<(n); _i++){ \
            FNumber _tmp=0; \
            _Pragma("acc loop independent, reduction(+:_tmp)") \
            for(size_t _j=0; _j<(m); _j++) \
                _tmp+=a[_i+(n)*_j]*v[_j]; \
            r[_i]=_tmp; \
        } \
    }

#if 0
static __inline__ void matmatmult(FNumber *r, 
                                  const FNumber *a, 
                                  const FNumber *b, 
                                  FNumber rw,
                                  size_t n, size_t k, size_t m)
{
    for(size_t i=0; i<n; i++){
        for(size_t j=0; j<m; j++){
            FNumber tmp=0;
            size_t kj=k*j;
            for(size_t l=0; l<k; l++)
                tmp+=a[i+n*l]*b[l+kj];
            if(rw==0)
                r[i+m*j]=tmp;
            else
                r[i+m*j]=rw*r[i+m*j]+tmp;
        }
    }
}
#endif

#endif /* _MISC_H */
