/*
 * dticode/src/normalise.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */

#include <string.h>
#include <stdio.h>
#include <math.h>

#include "sym2multi3.h"
#include "tensor.h"
#include "tensornorm.h"
#include "misc.h"
#include "acc_support.h"
#include "number.h"

static __inline__ FNumber mkfact_normalise(FNumber n, FNumber alpha)
{
    return (n==0 ? 1 : alpha/n);
}

static __inline__ FNumber mkfact_project(FNumber n, FNumber alpha)
{
    return alpha/fmax(alpha, n);
}



static __inline__ FNumber mkfact_threshold(FNumber y, FNumber ul, FNumber ur, FNumber sigma)
{
	
        float res=0.0;	
	if (y>sigma*ur)
		res=y-sigma*ur;
	if (y<sigma*ul){	        
		res=y-sigma*ul;}
	return res;
}

static __inline__ FNumber mkfact_threshold_prod(FNumber y, FNumber ul, FNumber ur)
{
	
        FNumber res;
      
	if (y<0)
		res=-y*ul;
	else
		res=-y*ur;
	//printf("y=%f\n",y);
	//printf("res=%f\n",res);
	
	return res;
}
static __inline__ FNumber mkfact_threshold_norm(FNumber y, FNumber ul, FNumber ur)
{
	
        FNumber res=0;
      
	if (y<ul)
		res=(y-ul)*(y-ul);
	if (y>ur)
		res=(y-ur)*(y-ur);

	
	return res;
}


/* Major hack to work around PGCC's 
 *  * Broken handling of _Pragma in macros (converted to #pragma 
 *    without linefeed thereafter), 
 *  * Failure to inline functions passed by pointers.
 */
#define MKFACT mkfact_normalise
#define APPLY_WEIGHT 1
#define WORKER do_normalise_w
#include "normalise_worker_template.h"

#define MKFACT mkfact_normalise
#define APPLY_WEIGHT 0
#define WORKER do_normalise_nw
#include "normalise_worker_template.h"

#define MKFACT mkfact_project
#define APPLY_WEIGHT 1
#define WORKER do_project_w
#include "normalise_worker_template.h"

#define MKFACT mkfact_project
#define APPLY_WEIGHT 0
#define WORKER do_project_nw
#include "normalise_worker_template.h"

#define MKFACT mkfact_threshold
#define APPLY_WEIGHT 1
#define WORKER do_threshold_simple
#include "threshold_worker_template.h"

#define MKFACT mkfact_threshold_prod
#define WORKER do_threshold_prod_simple
#include "threshold_prod_worker_template.h"

#define MKFACT mkfact_threshold_norm
#define WORKER do_threshold_norm_simple
#include "threshold_prod_worker_template.h"

static void do_normalise(size_t dim, size_t n, FNumber *p,
                         FNumber alpha, const FNumber *weight)
{
    if(weight)
        do_normalise_w(dim, n, p, alpha, weight);
    else
        do_normalise_nw(dim, n, p, alpha, weight);
}

static void do_project(size_t dim, size_t n, FNumber *p,
                       FNumber alpha, const FNumber *weight)
{
    if(weight)
        do_project_w(dim, n, p, alpha, weight);
    else
        do_project_nw(dim, n, p, alpha, weight);
}

static void do_threshold(size_t dim, size_t n, FNumber *p, FNumber *ulp, FNumber *urp, FNumber sigma)
{
        do_threshold_simple(dim, n, p, ulp, urp, sigma);
}
static FNumber do_threshold_prod(size_t dim, size_t n, FNumber *p, FNumber *ulp, FNumber *urp)
{
        FNumber sum=do_threshold_prod_simple(dim, n, p, ulp, urp);
return sum;
}

static FNumber do_threshold_norm(size_t dim, size_t n, FNumber *p, FNumber *ulp, FNumber *urp)
{
        FNumber sum=do_threshold_norm_simple(dim, n, p, ulp, urp);
return sum;
}


#ifndef MEX
void tensorfield_normalise_frob2(TensorField *t, FNumber alpha)
{
    do_normalise(t->elemsize, t->nelem, t->p, alpha,
                 tensorfield_get_weight(t));
}


void tensorfield_project_frob2(TensorField *t, FNumber alpha)
{
    if(alpha!=INFINITY){
        do_project(t->elemsize, t->nelem, t->p, alpha,
                   tensorfield_get_weight(t));
    }
}


void tensorfield_threshold_frob2(TensorField *t, TensorField *ul, TensorField *ur, FNumber sigma)
{
    
do_threshold(t->elemsize, t->nelem, t->p, ul->p, ur->p, sigma);
}

void tensorfield_threshold_prod(TensorField *t,TensorField *ul, TensorField *ur,FNumber *n)
{

FNumber sum=do_threshold_prod(t->elemsize, t->nelem, t->p, ul->p, ur->p );
//printf("sum=%f\n",sum);
*n=sum;
}

void tensorfield_threshold_norm(TensorField *t,TensorField *ul, TensorField *ur,FNumber *n)
{

FNumber sum=do_threshold_norm(t->elemsize, t->nelem, t->p, ul->p, ur->p );
//printf("sum=%f\n",sum);
*n=sqrt(sum);
}


#endif
