/*
 * dticode/tensorreg/fid_fourier_l2.h
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */


#ifndef _FID_FOURIER_L2_H
#define _FID_FOURIER_L2_H

#include "tensor.h"
#include "combine.h"
#include "matrix.h"
#include "fourier_sample.h"

typedef TensorField FidFourierL2Primal;
typedef TensorField FidFourierL2Dual;

typedef struct{
    const TensorField *f;
    const TensorField *f_orig;
    TensorField *f_tmp;
    bool full_f;
    const TensorField *mask;
    FNumber mask_term_weight;
    Sig sig;
    size_t n_samples;
    const FourierSample *samples;
    const FourierSample *samples_orig;
    FourierSample *samples_tmp;
    FNumber pdgap0_ab;
    FNumber pdgap0_c;
    FNumber ubound;
    FNumber fu_infty_bound2;
    TensorField *tmp;
} FidFourierL2Params;

void fid_fourier_l2_init(const TensorField *f,
                         const TensorField *mask,
                         FNumber mask_term_weight,
                         const Sig *sig,
                         size_t n_samples, 
                         const FourierSample *samples,
                         Fid **fid, 
                         FidFourierL2Params *param);

FNumber fid_fourier_l2_baseshift(FidFourierL2Primal *u, FidBaseshiftMode mode, 
                                 FidFourierL2Params *fp);

void fid_fourier_l2_init_primal(FidFourierL2Primal *u, FidInitPrimalStyle initstyle, 
                                FidFourierL2Params *fp);

bool fid_fourier_l2_grad(TensorField *grad, const TensorField *u, 
                         FidFourierL2Params *fp);

#endif /* _FID_FOURIER_L2_H */
