/*
 * dticode/src/parallel.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */
 
#include <stdio.h>
#include "parallel.h"
#include "timing.h"

#ifndef ACC_DEVICE
#define ACC_DEVICE acc_device_nvidia
#endif

void parallel_init1(int *argc, char **argv[])
{
#if _MPI
    MPI_Init(argc, argv);
#endif
}

void parallel_init2(/*int threads*/)
{
#if _OPENMP
    /*
    if(threads>0)
        omp_set_num_threads(threads);
    */
#endif

#if _OPENACC
#if _MPI
    int rank=0, numtasks=0;
    
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &numtasks);

    int numdev=acc_get_num_devices(ACC_DEVICE);
    /* Try to use device number mod(rank, numdev).
     * This should give every process on the same compute
     * node a different GPU, if the processes on the node
     * have consequtive ranks, and the number of processes
     * on the node does not exceed the number of GPUs.
     */
    acc_set_device_num(rank % numdev, ACC_DEVICE);
    acc_init(ACC_DEVICE);
    if(cpockp.verbose){
        char proc[MPI_MAX_PROCESSOR_NAME];
        int len;
        MPI_Get_processor_name(proc, &len);
        printf("Rank %d (%s) -> GPU %d\n", rank, proc, acc_get_device_num(ACC_DEVICE));
        fflush(stdout);
    }
#else
    acc_init(ACC_DEVICE);
#endif
#endif
}

void parallel_timing_print()
{
#if _MPI
    int rank=0, numtasks=0;
    
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &numtasks);

    char prefix[16];
    snprintf(prefix, 16, "Rank %d ", rank);
    if(rank!=0){
        /* Sequential print */
        int nothing;
        MPI_Recv(&nothing, 1, MPI_INT, rank-1, rank, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
    }
    timing_print(prefix);
    if(rank!=numtasks-1){
        /* Sequential print */
        int nothing;
        MPI_Send(&nothing, 1, MPI_INT, rank+1, rank+1, MPI_COMM_WORLD);
    }
#else
    timing_print(NULL);
#endif
}

void parallel_finalise()
{
#if _MPI
    MPI_Finalize();
#endif
}
