/*
 * dticode/src/posdef.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */

#include <string.h>
#include <libtu/output.h>

#include "sym2multi3.h"
#include "sym2multi2.h"
#include "sylvester.h"
#include "tensor.h"
#include "acc_support.h"
#include "number.h"

/* Major hack to work around PGCC's 
 *  * Broken handling of _Pragma in macros (converted to #pragma 
 *    without linefeed thereafter), 
 *  * Failure to inline functions passed by pointers.
 */
#define SYLVESTER sylvester_1
#define WORKER do_posdef1
#include "posdef_worker_template.h"

#define SYLVESTER sylvester_2
#define WORKER do_posdef2
#include "posdef_worker_template.h"

#define SYLVESTER sylvester_3
#define WORKER do_posdef3
#include "posdef_worker_template.h"


static char *do_posdef(size_t dim, size_t n, 
                       FNumber * DEVPTR a, char * pd,
                       FNumber threshold, int *res)
{    
    if(dim==1){
        *res=do_posdef1(1, n, a, pd, threshold);
    }else if(dim==I2_N){
        *res=do_posdef2(I2_N, n, a, pd, threshold);
    }else if(dim==I_N){
        *res=do_posdef3(I_N, n, a, pd, threshold);
    }else{
        return "Erroneous tensor";
    }
    
    return NULL;
}

#ifndef MEX
int tensorfield_posdef(const TensorField *a, FNumber threshold)
{
    char *err;
    int res=0;
    
    if(!a->sig.symmetric || a->sig.is_complex)
        die("tensorfield_posdef only implemented for symmetric real tensors");

    err=do_posdef(a->elemsize, a->nelem, a->p, NULL, threshold, &res);
    
    if(err!=NULL)
        die(err);
        
    return res;
}
#endif


