/*
 * dticode/src/reg.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <stdlib.h>
#include <string.h>
#include <libtu/output.h>
#include <libtu/misc.h>
#include "reg.h"
#include "reg_tv.h"
#include "reg_tgv2.h"
#include "reg_none.h"

void scan_regulariser(char **regname, FNumber *alpha, FNumber *beta, const char *param)
{
    *regname=ALLOC_N(char, strlen(param)+1);
    if(*regname==NULL)
        die_err();

#if USE_FLOAT
    size_t n=sscanf(param, "%[^,],%f,%f", *regname, alpha, beta);
#else
    size_t n=sscanf(param, "%[^,],%lf,%lf", *regname, alpha, beta);
#endif
    if(n<1)
        die("Invalid regulariser specification %s\n", param);
}

void init_regulariser(Reg **reg, RegP *regp, const char *regname_param, 
                      const Sig *sig, const char *diffalg)
{
    char *regname=NULL;
    FNumber alpha=0, beta=0;
    
    scan_regulariser(&regname, &alpha, &beta, regname_param);

    if(strcmp(regname, "td")==0){
        reg_tv_init(sig, TRUE, alpha, diffalg, reg, &regp->tv);
    }else if(strcmp(regname, "tv")==0){
        reg_tv_init(sig, FALSE, alpha, diffalg, reg, &regp->tv);
    }else if(strcmp(regname, "tgv2")==0){
        reg_tgv2_init(sig, alpha, beta, diffalg, reg, &regp->tgv2);
    }else if(strcmp(regname, "none")==0){
        reg_none_init(reg);
    }else{
        die("Invalid regularisation rule: %s", regname);
    }
    
    free(regname);
}
