/*
 * dticode/src/reg_none.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <libtu/output.h>
#include "vector.h"
#include "combine.h"
#include "reg_none.h"

static void none_k(RegNoneDual *a, const CombinedPrimal *b, 
                   bool addto_a, RegNoneParams *param)
{
}

static void none_kc(CombinedPrimal *a, const RegNoneDual *b, 
                    bool addto_au, bool addto_aw, RegNoneParams *param)
{
    if(!addto_au)
        vector_zero(COMBINED_U(a));
}

static void none_fcres(RegNoneDual *a, FNumber tau, RegNoneParams *param)
{
}


static void none_pdgap(FNumber *pdgap, 
                       const CombinedPrimal *x, 
                       const RegNoneDual *phi, /* y */
                       const CombinedPrimal *kcy, 
                       const RegNoneDual *eu,  /* kx */
                       CombinedPrimal *xwork,
                       unsigned int iter,
                       RegNoneParams *param)
{
    pdgap[0]=0;
    pdgap[1]=0;
}

static void none_init_dual_extra(RegNoneExtraPrimal *w, RegNoneDual *psi, 
                                 const RegNonePrimal *u, RegNoneParams *param)
{
}

static FNumber none_estimate(Estimate estim, RegNoneParams *param)
{
    if(estim==EstimateKNorm){
        return 0;
    }else if(estim==EstimateGUniformConvexity){
        return 0;
    }else if(estim==EstimateFCUniformConvexity){
        return 0;
    }else{
        return 0;
    }
}

static Reg reg_none={
    (RegK*)none_k,
    (RegKC*)none_kc,
    (RegFCres*)none_fcres,
    (RegGres*)noop,
    (RegPDgap*)none_pdgap,
    (RegAllocPrimal*)alloc_dummyvector,
    (RegAllocDual*)alloc_dummyvector,
    (RegInitDualExtra*)none_init_dual_extra,
    (RegEstimate*)none_estimate,
};

void reg_none_init(Reg **reg)
{
    *reg=&reg_none;
}
