/*
 * dticode/src/scaleplus.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */


#include <string.h>
#include <stdio.h>
#include <stddef.h>
#include <math.h>
#include <libtu/output.h>

#include "tensor.h"
#include "acc_support.h"
#include "number.h"

/*
function res=scaleplus(y, f, gamma, tau, sigma)
    res=(gamma*y+f*tau)/sigma;
end
*/

static __inline__ FNumber op_minus(FNumber y, FNumber f, FNumber gamma, FNumber tau)
{
    return y-f;
}

static __inline__ FNumber op_plus(FNumber y, FNumber f, FNumber gamma, FNumber tau)
{
    return y+f;
}

static __inline__ FNumber op_tauonly(FNumber y, FNumber f, FNumber gamma, FNumber tau)
{
    return y+tau*f;
}

static __inline__ FNumber op_full(FNumber y, FNumber f, FNumber gamma, FNumber tau)
{
    return gamma*y+tau*f;
}

/* Major hack to work around PGCC's 
 *  * Broken handling of _Pragma in macros (converted to #pragma 
 *    without linefeed thereafter), 
 *  * Failure to inline functions passed by pointers.
 */
#define WORKER worker_minus
#define OP op_minus
#include "scaleplus_worker_template.h"

#define WORKER worker_plus
#define OP op_plus
#include "scaleplus_worker_template.h"

#define WORKER worker_tauonly
#define OP op_tauonly
#include "scaleplus_worker_template.h"

#define WORKER worker_full
#define OP op_full
#include "scaleplus_worker_template.h"

typedef void Worker(size_t n, FNumber *res,
                    const FNumber *y, const FNumber *f,
                    FNumber gamma, FNumber tau);


static void do_scaleplus(size_t n, FNumber * DEVPTR res,
                         const FNumber * DEVPTR y, const FNumber * DEVPTR f,
                         FNumber gamma, FNumber tau, FNumber sigma)
{
    Worker *worker;
    
    if(sigma==1){
        if(gamma==1){
            if(tau==-1){
                worker=worker_minus;
            }else if(tau==1){
                worker=worker_plus;
            }else{
                worker=worker_tauonly;
            }
        }else{
            worker=worker_full;
        }
    }else{
        gamma/=sigma;
        tau/=sigma;
        worker=worker_full;
    }
    
    worker(n, res, y, f, gamma, tau);
}


#ifndef MEX
/* NOTE: <a> is allowed to equal <b> or <c> */
void tensorfield_scaleplus(TensorField* a, 
                           const TensorField* b,
                           const TensorField* c,
                           FNumber gamma, FNumber tau, FNumber sigma)
{
    sig_check_match(&(a->sig), &(b->sig));
    sig_check_match(&(b->sig), &(c->sig));
    
    do_scaleplus(a->nelem*a->elemsize, a->p, b->p, c->p, gamma, tau, sigma);
}
#endif

void tensorfield_scale(const TensorField *a, FNumber alpha)
{
    FNumber *p=a->p;
    const size_t n=a->nelem*a->elemsize;
    
    #pragma omp parallel for
    #pragma acc kernels loop deviceptr(p), independent
    for(size_t i=0; i<n; i++)
        p[i]=p[i]*alpha;
}
