/*
 * dticode/src/scaleproj.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */

#include <stddef.h>
#include <libtu/output.h>
#include <assert.h>
#include <stdio.h>

#include "scaleproj.h"
#include "number.h"
#include "tensor.h"

typedef void ScaleProj(FNumber *v, size_t count, const FNumber *m_mat, 
                       FNumber minm, FNumber sigma, FNumber gamma,
                       int iters, FNumber rho);

extern ScaleProj scaleproj_1;
extern ScaleProj scaleproj_2;
extern ScaleProj scaleproj_3;

// Solve M t + N_{>=0}(t) = M t_0, t>=0,
// where t_0 is the initial (parameter) value of t.
void tensorfield_scaleproj(TensorField *t, 
                           const Matrix *mmat, 
                           FNumber minm, 
                           FNumber sigma, 
                           FNumber gamma, 
                           int iters, 
                           FNumber rho)
{
    const size_t dim=t->sig.dim;
    ScaleProj *scaleproj=NULL;
    
    assert(t->sig.order==2 && t->sig.symmetric && !t->sig.is_complex);
    assert(mmat->m==t->elemsize && mmat->n==t->elemsize);

    if(dim==1)
        scaleproj=scaleproj_1;
    else if(dim==2)
        scaleproj=scaleproj_2;
    else if(dim==3)
        scaleproj=scaleproj_3;
    else
        //die("scaleproj: unimplemented dimension");
	printf("scaleproj: unimplemented dimension");
        
    scaleproj(t->p, t->nelem, mmat->p, 
              minm, sigma, gamma, iters, rho);
}

