/*
 * dticode/src/scaleproj_lmat.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <libtu/output.h>

#include "scaleproj_symmat.h"
#include "scaleproj_lmat.h"

static void do_write_lmat(const char *pn, 
                          const char *lpformat, 
                          const char *lineend,
                          int dim, 
                          const SymMatLUT *lut)
{
    const size_t n=lut->lmat_lut_size;
    const size_t lms=lut->lmat_size;
    
    for(size_t j=0; j<lms; j++){
        LMatLUT *llut=lut->lmat_lut;
        ssize_t first=-1;
        ssize_t second=-1;
        for(size_t i=0; i<n; i++){
            if(llut[i].lp_index==j){
                assert(first==-1 || second==-1);
                if(first==-1)
                    first=llut[i].p_index;
                else
                    second=llut[i].p_index;
            }
        }
        printf(lpformat, j);
        if(first==-1)
            printf("0;");
        else if(second==-1)
            printf("%s[%ld]*FNUMBER(0.5);", pn, first);
        else if(first==second)
            printf("%s[%ld];", pn, first);
        else
            printf("(%s[%ld]+%s[%ld])*FNUMBER(0.5);", pn, first, pn, second);
        printf(lineend);
    }
}

static void write_make_lmat(int dim, const SymMatLUT *lut)
{
    
    printf("static __inline__ void make_lmat_%d(FNumber *lp, const FNumber *p)\n", dim);
    printf("{\n");
    do_write_lmat("p", "    lp[%ld] = ", "\n", dim, lut);
    printf("}\n\n");
}

static void write_make_ld_lpm(int dim, const SymMatLUT *lut)
{
    const size_t sz=symsize(dim);
    const char *lineend=" \\\n";
    
    //printf("static __inline__ void make_ld_lpm%d(FNumber *resmat, const FNumber *p, const FNumber *d, const FNumber *mmat)\n", dim);
    printf("#define MAKE_LD_LPM_%d(resmat, p, d, mmat)%s", dim, lineend);
    printf("{%s", lineend);
    do_write_lmat("p", "    const FNumber lp_%ld = ", lineend, dim, lut);
    do_write_lmat("d", "    const FNumber ld_%ld = ", lineend, dim, lut);
    for(size_t i=0; i<sz; i++){
        for(size_t j=0; j<sz; j++){
            printf("    resmat[%ld] = ld_%ld", i+sz*j, i+sz*j);
            
            for(size_t l=0; l<sz; l++)
                printf(" + lp_%ld*mmat[%ld]", i+sz*l, l+sz*j);
                
            printf(";%s", lineend);
        }
    }

    printf("}\n\n");
}

int main(int argc, char *argv[])
{
    SymMatLUT *lut;
    int dim;
    
    if(argc!=2)
        //die("Invalid argumens");
        return 0;
    
    dim=atoi(argv[1]);
    if(dim<=0)
        //die("Invalid dimension");
	return 0;
        
    lut=symmat_lut_create(dim);
    
    printf("#define LMAT_SIZE_%d %ld\n\n", dim, lut->lmat_size);
    write_make_lmat(dim, lut);
    write_make_ld_lpm(dim, lut);
    
    symmat_lut_free(lut);
    
    return 0;
}
