/*
 * dticode/src/scaleproj_lmat.h
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */

#ifndef _SCALEPROJ_LMAT_H
#define _SCALEPROJ_LMAT_H
 
#include "scaleproj_symmat.h"
 
typedef struct{
    size_t lp_index;
    size_t p_index;
} LMatLUT;

typedef struct{
    size_t dim;
    size_t sz;
    /* L(p) */
    size_t lmat_size;
    size_t lmat_lut_size;
    LMatLUT *lmat_lut;
} SymMatLUT;

static __inline__ size_t index_sym_colstart(size_t j)
{
    return (j+1)*j/2;
}

static __inline__ size_t index_sym(size_t i, size_t j)
{
    if(i<=j)
        return index_sym_colstart(j)+i;
    else
        return index_sym_colstart(i)+j;
}

void make_lmat(double *lp, const double *p, size_t dim)
{
    size_t i, j, k, ji, jk, ki, sz;
    /* lp(x)=(px + xp)/2
     * ==> lp(x)_ji = sum_k (p_jk x_ki + x_jk p_ki)/2
     */
    sz=symsize(dim);
    memset(lp, 0, sizeof(double)*sz*sz);
    for(i=0; i<dim; i++){
        ji=index_sym_colstart(i);
        for(j=0; j<=i; j++, ji++){
            /*ij=index_sym(i, j);*/
            for(k=0; k<dim; k++){
                jk=index_sym(j, k);
                ki=index_sym(k, i);
                lp[ji+sz*ki]+=p[jk]/2;
                lp[ji+sz*jk]+=p[ki]/2;
            }
        }
    }
}

static int calc_lmat_lut(SymMatLUT *lut, size_t dim)
{
    size_t idx=0;
    size_t i, j, k, ji, jk, ki;
    size_t sz=symsize(dim);
    size_t n=sz*dim*2;
    LMatLUT *llut=malloc(n*sizeof(LMatLUT));

    if(llut==NULL)
        return 0;

    lut->lmat_lut=llut;
    lut->lmat_lut_size=n;
    lut->lmat_size=sz*sz;
    
    /* lp(x)=(px + xp)/2
     * ==> lp(x)_ji = sum_k (p_jk x_ki + x_jk p_ki)/2
     */
    for(i=0; i<dim; i++){
        ji=index_sym_colstart(i);
        for(j=0; j<=i; j++, ji++){
            for(k=0; k<dim; k++){
                assert(idx+2<=n);
                //ji=index_sym(j, i);
                jk=index_sym(j, k);
                ki=index_sym(k, i);
                llut[idx].lp_index=ji+sz*ki;
                llut[idx].p_index=jk;
                idx++;
                llut[idx].lp_index=ji+sz*jk;
                llut[idx].p_index=ki;
                idx++;
            }
        }
    }
    assert(idx==n);
    
    return 1;
}

static __inline__ void make_lmat_precalc(double *lp, const double *p, 
                                         const SymMatLUT *lut)
{
    LMatLUT *llut=lut->lmat_lut;
    size_t n=lut->lmat_lut_size;
    size_t i;
    
    memset(lp, 0, sizeof(double)*lut->lmat_size);
    
    for(i=0; i<n; i++, llut++)
        lp[llut->lp_index]+=p[llut->p_index]/2;
}


static SymMatLUT *symmat_lut_create(size_t dim)
{
    SymMatLUT *lut=malloc(sizeof(SymMatLUT));
    
    if(lut==NULL)
        return NULL;
        
    lut->dim=dim;
    lut->sz=symsize(dim);
    
    if(!calc_lmat_lut(lut, dim)){
        free(lut);
        return NULL;
    }
    
    return lut;
}

static void symmat_lut_free(SymMatLUT *lut)
{
    free(lut->lmat_lut);
    free(lut);
}

#endif /* _SCALEPROJ_LMAT_H */

