/*
 * dticode/src/scaleproj_symmat.2.h
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */
 
#ifndef _SCALEPROJ_SYMMAT_2_H
#define _SCALEPROJ_SYMMAT_2_H

#include "number.h"
#include "sym2multi2.h"

/* */

#define symsize(dim) I2_N
#define dimrank(dim) 2

//
// Lots of macros because PGCC OpenACC implementation hates functions
//

//static __inline__ void make_eye(FNumber *p, FNumber scale, size_t dim)
#define make_eye(p, scale, dim) \
{ \
    p[I2_XX]=scale; \
    p[I2_YY]=scale; \
    p[I2_XY]=0; \
}

static __inline__ FNumber symtrace(const FNumber *p, size_t dim)
{
    return p[I2_XX]+p[I2_YY];
}

//static __inline__ void symproj_pe(FNumber *proj, const FNumber *p, size_t dim)
#define symproj_pe(proj, p, dim) \
{ \
    FNumber mu=symtrace(p, dim)/dimrank(dim); \
 \
    proj[I2_XX]=p[I2_XX]-mu; \
    proj[I2_YY]=p[I2_YY]-mu; \
    proj[I2_XY]=p[I2_XY]; \
}

static __inline__ FNumber symiprod(const FNumber *p, const FNumber *d, size_t dim)
{
    return p[I2_XX]*d[I2_XX]+p[I2_YY]*d[I2_YY]+2*p[I2_XY]*d[I2_XY];
}

static __inline__ FNumber symfrob(const FNumber *p, size_t dim)
{
    return sqrt(symiprod(p, p, dim));
}


//static __inline__ void symprod(FNumber *res, 
//                               const FNumber *p,
//                               const FNumber *d,
//                               size_t dim)
#define symprod(res, p, d, dim) \
{ \
    res[I2_XX]=p[I2_XX]*d[I2_XX]+p[I2_XY]*d[I2_XY]; \
    res[I2_YY]=p[I2_YY]*d[I2_YY]+p[I2_XY]*d[I2_XY]; \
    res[I2_XY]=(p[I2_XX]*d[I2_XY]+p[I2_XY]*d[I2_YY] \
                + p[I2_YY]*d[I2_XY]+p[I2_XY]*d[I2_XX])*FNUMBER(0.5); \
}

//static __inline__ void make_dirvec(FNumber *v, const FNumber *p, const FNumber *d, 
//                                   FNumber emult, size_t dim)
#define make_dirvec(v, p, d, emult, dim) \
{ \
    symprod(v, p, d, dim); \
 \
    v[I2_XX]=emult-v[I2_XX]; \
    v[I2_YY]=emult-v[I2_YY]; \
    v[I2_XY]=-v[I2_XY]; \
}

#endif /* _SCALEPROJ_SYMMAT_2_H */
