/*
 * dticode/src/scaleproj_symmat.2.h
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */
 
#ifndef _SCALEPROJ_SYMMAT_3_H
#define _SCALEPROJ_SYMMAT_3_H

#include "number.h"
#include "sym2multi3.h"

/* */

#define symsize(dim) I_N
#define dimrank(dim) 3

//
// Lots of macros because PGCC OpenACC implementation hates functions
//

//static __inline__ void make_eye(FNumber *p, FNumber scale, size_t dim)
#define make_eye(p, scale, dim) \
{ \
    p[I_XX]=scale; \
    p[I_YY]=scale; \
    p[I_ZZ]=scale; \
    p[I_XY]=0; \
    p[I_XZ]=0; \
    p[I_YZ]=0; \
}

static __inline__ FNumber symtrace(const FNumber *p, size_t dim)
{
    return p[I_XX]+p[I_YY]+p[I_ZZ];
}

//static __inline__ void symproj_pe(FNumber *proj, const FNumber *p, size_t dim)
#define symproj_pe(proj, p, dim) \
{ \
    FNumber mu=symtrace(p, dim)/dimrank(dim); \
 \
    proj[I_XX]=p[I_XX]-mu; \
    proj[I_YY]=p[I_YY]-mu; \
    proj[I_ZZ]=p[I_ZZ]-mu; \
    proj[I_XY]=p[I_XY]; \
    proj[I_XZ]=p[I_XZ]; \
    proj[I_YZ]=p[I_YZ]; \
}

static __inline__ FNumber symiprod(const FNumber *p, const FNumber *d, size_t dim)
{
    return p[I_XX]*d[I_XX]+p[I_YY]*d[I_YY]+p[I_ZZ]*d[I_ZZ]
           +2*(p[I_XY]*d[I_XY]+p[I_XZ]*d[I_XZ]+p[I_YZ]*d[I_YZ]);
}

static __inline__ FNumber symfrob(const FNumber *p, size_t dim)
{
    return sqrt(symiprod(p, p, dim));
}

 
//static __inline__ void symprod(FNumber *res, 
//                               const FNumber *p,
//                               const FNumber *d,
//                               size_t dim)
#define symprod(res, p, d, dim) \
{ \
    res[I_XX]=p[I_XX]*d[I_XX]+p[I_XY]*d[I_XY]+p[I_XZ]*d[I_XZ]; \
    res[I_YY]=p[I_XY]*d[I_XY]+p[I_YY]*d[I_YY]+p[I_YZ]*d[I_YZ]; \
    res[I_ZZ]=p[I_XZ]*d[I_XZ]+p[I_YZ]*d[I_YZ]+p[I_ZZ]*d[I_ZZ]; \
    res[I_XY]=( p[I_XX]*d[I_XY]+p[I_XY]*d[I_YY]+p[I_XZ]*d[I_YZ] \
               +d[I_XX]*p[I_XY]+d[I_XY]*p[I_YY]+d[I_XZ]*p[I_YZ])*FNUMBER(0.5); \
    res[I_XZ]=( p[I_XX]*d[I_XZ]+p[I_XY]*d[I_YZ]+p[I_XZ]*d[I_ZZ] \
               +d[I_XX]*p[I_XZ]+d[I_XY]*p[I_YZ]+d[I_XZ]*p[I_ZZ])*FNUMBER(0.5); \
    res[I_YZ]=( p[I_YX]*d[I_XZ]+p[I_YY]*d[I_YZ]+p[I_YZ]*d[I_ZZ] \
               +d[I_YX]*p[I_XZ]+d[I_YY]*p[I_YZ]+d[I_YZ]*p[I_ZZ])*FNUMBER(0.5); \
}

//static __inline__ void make_dirvec(FNumber *v, const FNumber *p, const FNumber *d, 
//                                   FNumber emult, size_t dim)
#define make_dirvec(v, p, d, emult, dim) \
{ \
    symprod(v, p, d, dim); \
 \
    v[I_XX]=emult-v[I_XX]; \
    v[I_YY]=emult-v[I_YY]; \
    v[I_ZZ]=emult-v[I_ZZ]; \
    v[I_XY]=-v[I_XY]; \
    v[I_XZ]=-v[I_XZ]; \
    v[I_YZ]=-v[I_YZ]; \
}

#endif /* _SCALEPROJ_SYMMAT_3_H */
