/*
 * dticode/src/stejskal-tanner.h
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */

#define BSIGN -

#define add_nothing(p, idx, x)
#define set_nothing(p, idx, x)

#define GETONE(p, ii) 1.0

#define NULLTF ((TensorField*)NULL)

#define DTI_CALC(y, s0, x, z, bmat, sety, gets0, getx, getz, C) { \
        size_t nelem=y->nelem; \
        FNumber * DEVPTR yp=y->p; \
        const FNumber * DEVPTR xp=x->p; \
        const FNumber * DEVPTR bp=bmat->p; \
        const FNumber * DEVPTR s0p=(s0 ? s0->p : NULL); \
        const FNumber * DEVPTR zp=(z ? z->p : NULL); \
        const size_t n=bmat->n; \
        const size_t m=bmat->m; \
        assert(x->elemsize==m); \
        assert(!s0 || (y->nelem==x->nelem && s0->nelem==y->nelem)); \
        /*assert(y->elemsize==C*n);*/ \
        assert(!s0 || (s0->elemsize==C*n)); \
        /* Calculate y := s_0  exp(y) = s_0 exp(<b,u>) */ \
        _Pragma("omp parallel for") \
        _Pragma("acc kernels loop deviceptr(yp, sp, xp, s0p), copyin(bp[0:n*m]), independent") \
        for(size_t k=0; k<nelem; k++){ \
            size_t ypoff=k*n; \
            size_t xpoff=k*m; \
            _Pragma("acc loop vector, independent") \
            for(size_t i=0; i<n; i++){ \
                FNumber bx=0; \
                for(size_t j=0; j<m; j++) \
                    bx+=bp[i+n*j]*getx(xp, xpoff+j); \
                sety(yp, ypoff+i, gets0(s0p, ypoff+i)*getz(zp, k)*fn_exp(BSIGN bx)); \
            } \
        } \
    }

#define DTI_CALC_A(y, s0, x, z, bmat, addto_y, yset_type, gets0, getx, getz, C) { \
        if(addto_y){ \
            DTI_CALC(y, s0, x, z, bmat, add_##yset_type, gets0, getx, getz, C); \
        }else{ \
            DTI_CALC(y, s0, x, z, bmat, set_##yset_type, gets0, getx, getz, C); \
        } \
    }

#define DTI_APPLY_GRAD(y, s0, x0, z0, x, z, bmat, sety, gets0, getx, getz, C) { \
        size_t nelem=y->nelem; \
        FNumber * DEVPTR yp=y->p; \
        const FNumber * DEVPTR xp=x->p; \
        const FNumber * DEVPTR x0p=x0->p; \
        const FNumber * DEVPTR bp=bmat->p; \
        const FNumber * DEVPTR s0p=(s0 ? s0->p : NULL); \
        const FNumber * DEVPTR z0p=(z0 ? z0->p : NULL); \
        const FNumber * DEVPTR zp=(z ? z->p : NULL); \
        const size_t n=bmat->n; \
        const size_t m=bmat->m; \
        assert(x->elemsize==m && x0->elemsize==m); \
        assert(y->nelem==x->nelem && x0->nelem==x->nelem && (!s0 || s0->nelem==x->nelem)); \
        assert(y->elemsize==C*n); \
        assert(!s0 || s0->elemsize==C*n); \
        _Pragma("acc kernels loop deviceptr(yp, xp, zp, x0p, z0p, s0p, bp), copyin(bp[0:n*m]), independent") \
        _Pragma("omp parallel for") \
        for(size_t k=0; k<nelem; k++){ \
            size_t ypoff=k*n; \
            size_t xpoff=k*m; \
            _Pragma("acc loop vector, independent") \
            for(size_t i=0; i<n; i++){ \
                FNumber bx=0, bx0=0; \
                for(size_t j=0; j<m; j++){ \
                    bx+=bp[i+n*j]*getx(xp, xpoff+j); \
                    bx0+=bp[i+n*j]*getx(x0p, xpoff+j); \
                } \
                sety(yp, ypoff+i, (gets0(s0p, ypoff+i)*(BSIGN getz(z0p, k)*bx+getz(zp, k))*fn_exp(BSIGN bx0))); \
            } \
        } \
    }


#define DTI_APPLY_GRAD_A(y, s0, x, z, x0, z0, bmat, addto_y, yset_type, gets0, getx, getz, C) { \
        if(addto_y){ \
            DTI_APPLY_GRAD(y, s0, x, z, x0, z0, bmat, add_##yset_type, gets0, getx, getz, C); \
        }else{ \
            DTI_APPLY_GRAD(y, s0, x, z, x0, z0, bmat, set_##yset_type, gets0, getx, getz, C); \
        } \
    }

#define DTI_APPLY_GRADC(x, z, s0, y, x0, z0, bmat, addx, addz, gets0, getx, getz, gety, C) { \
        size_t nelem=y->nelem; \
        const FNumber * DEVPTR yp=y->p; \
        FNumber * DEVPTR xp=x->p; \
        const FNumber * DEVPTR x0p=x0->p; \
        const FNumber * DEVPTR z0p=(z0 ? z0->p : NULL); \
        const FNumber * DEVPTR bp=bmat->p; \
        const FNumber * DEVPTR s0p=(s0 ? s0->p : NULL); \
        const size_t n=bmat->n; \
        const size_t m=bmat->m; \
        assert(x->elemsize==m && x0->elemsize==m); \
        assert(y->nelem==x->nelem && x0->nelem==x->nelem && (!s0 || s0->nelem==x->nelem)); \
        assert(y->elemsize==C*n); \
        assert(!s0 || s0->elemsize==C*n); \
        _Pragma("acc kernels loop deviceptr(yp, xp, x0p, z0p, s0p, bp), copyin(bp[0:n*m]), independent") \
        _Pragma("omp parallel for") \
        /* Loop over voxels */ \
        for(size_t k=0; k<nelem; k++){ \
            size_t ypoff=k*n; \
            size_t xpoff=k*m; \
            _Pragma("acc loop vector, independent") \
            /* Loop over DWI signal indices */ \
            for(size_t i=0; i<n; i++){ \
                FNumber bx0=0; \
                /* Loop over tensor elements */ \
                for(size_t j=0; j<m; j++){ \
                    bx0+=bp[i+n*j]*getx(x0p, xpoff+j); \
                } \
                /* tmp = (<s_0 exp(-<b, u>), dy_i>, for i signal index */ \
                __typeof__(gets0(s0p, 0)) tmp0=gets0(s0p, ypoff+i)*gety(yp, ypoff+i)*fn_exp(BSIGN bx0); \
                __typeof__(tmp0*getz(z0p, 0)) tmp=tmp0*getz(z0p, k); \
                /* x -= b_i * <s_0 exp(-<b, u>, dy_i>, for i signal index */ \
                /* Loop over tensor elements. */ \
                for(size_t j=0; j<m; j++) \
                    addx(xp, xpoff+j, BSIGN bp[i+n*j]*tmp); \
                addz(zp, k, tmp0); \
            } \
        } \
    }

#define DTI_APPLY_GRADC_A(x, z, s0, y, x0, z0, bmat, addto_x, xset_type, zset_type, gets0, getx, getz, gety, C) { \
        if(!addto_x){ \
            tensorfield_zero(x); \
            if(z) \
                tensorfield_zero(z); \
        } \
        DTI_APPLY_GRADC(x, z, s0, y, x0, z0, bmat, add_##xset_type, add_##zset_type, gets0, getx, getz, gety, C); \
    }

#define DO_YES(X) X
#define DO_NO(X) /* */

#define DTI_EST_GRAD_NORM(s0expmax, s0expmaxnz, s0, x0, z0, bmat, gets0, getx, getz, abs2, C, DONZ) { \
        size_t nelem=x0->nelem; \
        const FNumber * DEVPTR x0p=x0->p; \
        const FNumber * DEVPTR bp=bmat->p; \
        const FNumber * DEVPTR s0p=(s0 ? s0->p : NULL); \
        const FNumber * DEVPTR z0p=(z0 ? z0->p : NULL); \
        const size_t n=bmat->n; \
        const size_t m=bmat->m; \
        assert(x0->elemsize==m && x0->elemsize==m); \
        assert(!s0 || s0->nelem==x0->nelem); \
        assert(!s0 || s0->elemsize==C*n); \
        s0expmax=0; \
        _Pragma("acc kernels loop deviceptr(xp, s0p, bp), copyin(bp[0:n*m]), independent") \
        _Pragma("omp parallel for") \
        for(size_t k=0; k<nelem; k++){ \
            size_t ypoff=k*n; \
            size_t xpoff=k*m; \
            FNumber nrm=0; \
            FNumber nrmnz=0; \
            _Pragma("acc loop vector, independent") \
            for(size_t i=0; i<n; i++){ \
                FNumber bx0=0; \
                for(size_t j=0; j<m; j++) \
                    bx0+=bp[i+n*j]*getx(x0p, xpoff+j); \
                __typeof__(gets0(s0p, ypoff+i)) s0ex=gets0(s0p, ypoff+i)*fn_exp(BSIGN bx0); \
                nrm+=abs2(getz(z0p, k)*s0ex); \
                DONZ(nrmnz+=abs2(s0ex)); \
            } \
            s0expmax=fn_max(s0expmax, fn_sqrt(nrm)); \
            DONZ(s0expmaxnz=fn_max(s0expmaxnz, fn_sqrt(nrmnz))); \
        } \
    }

