/*
 * dticode/src/sumdot1.c
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */

#include <string.h>
#include <stdio.h>
#include <math.h>
#include <libtu/output.h>

#include "sym2multi3.h"
#include "tensor.h"
#include "acc_support.h"
#include "number.h"
#include "misc.h"



/* Major hack to work around PGCC's 
 *  * Broken handling of _Pragma in macros (converted to #pragma 
 *    without linefeed thereafter), 
 *  * Failure to inline functions passed by pointers.
 */

#define UPDATE UPDATE_PLUS
#define POWER fn_sqrt
#define DIFF powerdiff
#define APPLY_WEIGHT 1
#define WORKER sumdot1_sqrt_diff_w
#include "sumdot1_worker_template.h"

#define UPDATE UPDATE_PLUS
#define POWER fn_sqrt
#define DIFF powerdiff
#define APPLY_WEIGHT 0
#define WORKER sumdot1_sqrt_diff_nw
#include "sumdot1_worker_template.h"

#define UPDATE UPDATE_PLUS
#define POWER fn_sqrt
#define DIFF mult
#define APPLY_WEIGHT 1
#define WORKER sumdot1_sqrt_mult_w
#include "sumdot1_worker_template.h"

#define UPDATE UPDATE_PLUS
#define POWER fn_sqrt
#define DIFF mult
#define APPLY_WEIGHT 0
#define WORKER sumdot1_sqrt_mult_nw
#include "sumdot1_worker_template.h"

#define UPDATE UPDATE_PLUS
#define POWER idop
#define DIFF powerdiff
#define APPLY_WEIGHT 1
#define WORKER sumdot1_diff_w
#include "sumdot1_worker_template.h"

#define UPDATE UPDATE_PLUS
#define POWER idop
#define DIFF powerdiff
#define APPLY_WEIGHT 0
#define WORKER sumdot1_diff_nw
#include "sumdot1_worker_template.h"

#define UPDATE UPDATE_PLUS
#define POWER idop
#define DIFF mult
#define APPLY_WEIGHT 1
#define WORKER sumdot1_mult_w
#include "sumdot1_worker_template.h"

#define UPDATE UPDATE_PLUS
#define POWER idop
#define DIFF mult
#define APPLY_WEIGHT 0
#define WORKER sumdot1_mult_nw
#include "sumdot1_worker_template.h"

/* */

#define UPDATE UPDATE_MAX
#define POWER fn_sqrt
#define DIFF powerdiff
#define APPLY_WEIGHT 1
#define WORKER infdot1_sqrt_diff_w
#include "sumdot1_worker_template.h"

#define UPDATE UPDATE_MAX
#define POWER fn_sqrt
#define DIFF powerdiff
#define APPLY_WEIGHT 0
#define WORKER infdot1_sqrt_diff_nw
#include "sumdot1_worker_template.h"

#define UPDATE UPDATE_MAX
#define POWER fn_sqrt
#define DIFF mult
#define APPLY_WEIGHT 1
#define WORKER infdot1_sqrt_mult_w
#include "sumdot1_worker_template.h"

#define UPDATE UPDATE_MAX
#define POWER fn_sqrt
#define DIFF mult
#define APPLY_WEIGHT 0
#define WORKER infdot1_sqrt_mult_nw
#include "sumdot1_worker_template.h"

#define UPDATE UPDATE_MAX
#define POWER idop
#define DIFF powerdiff
#define APPLY_WEIGHT 1
#define WORKER infdot1_diff_w
#include "sumdot1_worker_template.h"

#define UPDATE UPDATE_MAX
#define POWER idop
#define DIFF powerdiff
#define APPLY_WEIGHT 0
#define WORKER infdot1_diff_nw
#include "sumdot1_worker_template.h"

#define UPDATE UPDATE_MAX
#define POWER idop
#define DIFF mult
#define APPLY_WEIGHT 1
#define WORKER infdot1_mult_w
#include "sumdot1_worker_template.h"

#define UPDATE UPDATE_MAX
#define POWER idop
#define DIFF mult
#define APPLY_WEIGHT 0
#define WORKER infdot1_mult_nw
#include "sumdot1_worker_template.h"


typedef FNumber Worker(size_t dim, size_t n, 
                      const FNumber * DEVPTR a, const FNumber * DEVPTR b, 
                      const FNumber * DEVPTR weight);

static char *do_sumdot1(size_t dim, size_t n, 
                        const FNumber * DEVPTR a, 
                        const FNumber * DEVPTR b,
                        const FNumber * DEVPTR weight,
                        int diff, int power, 
                        FNumber *res)
{    
    Worker *worker;
    
    if(power!=1 && power!=2)
        return "Unsupported power";
        
    if(diff){
        if(power==2){
            worker=(weight ? sumdot1_sqrt_diff_w : sumdot1_sqrt_diff_nw);
        }else{
            worker=(weight ? sumdot1_diff_w : sumdot1_diff_nw);
        }
    }else{
        if(power==2){
            worker=(weight ? sumdot1_sqrt_mult_w : sumdot1_sqrt_mult_nw);
        }else{
            worker=(weight ? sumdot1_mult_w : sumdot1_mult_nw);
        }
    }
    
    *res=worker(dim, n, a, b, weight);
    
    return NULL;                
}

/* */

FNumber tensorfield_sumdot1(const TensorField *a, const TensorField *b, 
                           int diff, int power)
{
    FNumber res=0;
    char *err;
    const FNumber *wm;
    
    sig_check_match(&(a->sig), &(b->sig));

    wm=tensorfield_get_weight(a);

    err=do_sumdot1(a->elemsize, a->nelem, a->p, b->p, wm, diff, power, &res);
    if(err!=NULL)
        die(err);
    
#if _MPI
    FNumber send=res;
    MPI_Allreduce(&send, &res, 1, FNUMBER_MPI_TYPE, MPI_SUM, MPI_COMM_WORLD);
#endif

    return res;
}


static char *do_infdot1(size_t dim, size_t n, 
                        const FNumber * DEVPTR a, 
                        const FNumber * DEVPTR b,
                        const FNumber * DEVPTR weight,
                        int diff, int power, 
                        FNumber *res)
{    
    Worker *worker;
    
    if(power!=1 && power!=2)
        return "Unsupported power";

    if(diff){
        if(power==2){
            worker=(weight ? infdot1_sqrt_diff_w : infdot1_sqrt_diff_nw);
        }else{
            worker=(weight ? infdot1_diff_w : infdot1_diff_nw);
        }
    }else{
        if(power==2){
            worker=(weight ? infdot1_sqrt_mult_w : infdot1_sqrt_mult_nw);
        }else{
            worker=(weight ? infdot1_mult_w : infdot1_mult_nw);
        }
    }

    *res=worker(dim, n, a, b, weight);
    
    return NULL;                
}


FNumber tensorfield_infdot1(const TensorField *a, const TensorField *b, 
                           int diff, int power)
{
    FNumber res=0;
    char *err;
    const FNumber *wm;
    
    sig_check_match(&(a->sig), &(b->sig));

    wm=tensorfield_get_weight(a);

    err=do_infdot1(a->elemsize, a->nelem, a->p, b->p, wm, diff, power, &res);
    if(err!=NULL)
        die(err);

#if _MPI
    FNumber send=res;
    MPI_Allreduce(&send, &res, 1, FNUMBER_MPI_TYPE, MPI_MAX, MPI_COMM_WORLD);
#endif    
    return res;
}


