/*
 * dticode/src/sumdot1_worker_template.h
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */

/* Major hack to work around PGCC's 
 *  * Broken handling of _Pragma in macros (converted to #pragma 
 *    without linefeed thereafter), 
 *  * Failure to inline functions passed by pointers.
 */

static FNumber WORKER(size_t dima, size_t dimb, size_t n, 
                      const FNumber * restrict DEVPTR a, 
                      const FNumber * restrict DEVPTR b, 
                      const FNumber * restrict mat,
                      const FNumber * restrict DEVPTR weight)
{
    FNumber v=0;

    #pragma omp parallel
#if APPLY_WEIGHT==1
    #pragma acc kernels deviceptr(a, b, weight), copyin(mat[0:dima*dimb])
#else
    #pragma acc kernels deviceptr(a, b), copyin(mat[0:dima*dimb])
#endif
    {
        FNumber locv=0;
        
#if UPDATE==UPDATE_PLUS
 #define REDUCTION +
 #define UPDATEX plus
#elif UPDATE==UPDATE_MAX
 #define REDUCTION max
 #define UPDATEX fn_max
#else
 #warning "unknown update"
#endif
 
        #pragma omp for
        #pragma acc loop gang, vector, independent, reduction(REDUCTION:locv)
        for(size_t i=0; i<n; i++){
            FNumber w=0;
            const size_t idima=i*dima;
            const size_t idimb=i*dimb;
            
            #pragma acc loop seq, independent, reduction(+:w)
            for(size_t j=0; j<dima; j++){
                FNumber tmp=0;
                for(size_t k=0; k<dimb; k++)
                    tmp+=mat[j+dima*k]*b[idimb+k];
#if APPLY_WEIGHT==1                    
                w+=DIFF(a[idima+j], tmp)*weight[j];
#else
                w+=DIFF(a[idima+j], tmp);
#endif
            }
                
            locv=UPDATEX(locv, POWER(w));
        }
        
        /* The next should be optimised out for OpenACC */
        #pragma omp critical
        v=UPDATEX(v, locv);
    }

    return v;
}

#undef UPDATEX
#undef REDUCTION
#undef WORKER
#undef APPLY_WEIGHT
#undef UPDATE
#undef DIFF
#undef POWER
