/*
 * dticode/src/tensor.h
 *
 * Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
 *
 */


#ifndef _TENSOR_H
#define _TENSOR_H

#if _MPI
#include "mpi.h"
#endif

#include <libtu/obj.h>

#include "lut.h"
#include "number.h"
#include "acc_support.h"
#include "vector.h"

typedef struct{
    int dim;
    int order;
    size_t w;
    size_t h;
    size_t d;
    bool symmetric;
    bool is_complex;
    FNumber xres;
    FNumber yres;
    FNumber zres;
} Sig;


#if _MPI
#define TENSORFIELD_FULLSIG(T) ((T)->fullsig)
#else
#define TENSORFIELD_FULLSIG(T) ((T)->sig)
#endif

INTRCLASS(TensorField);

DECLCLASS(TensorField){
    Vector vector;
    Sig sig;
    FNumber * DEVPTR p;
    size_t nelem;
    size_t elemsize;
#if _MPI
    Sig fullsig;
    FNumber * DEVPTR fullp;
    MPI_Datatype slicetype;
#if _OPENACC
    FNumber *transferbuf;
#endif
#endif
};

extern TensorField *alloc_tensorfield(const Sig *sig);
extern TensorField *load_tensorfield(const Sig *sig, const char *file);
extern void save_tensorfield(const TensorField *t, const char *file);
extern void free_tensorfield(TensorField *t);

extern void tensorfield_zero(TensorField *t);
extern void tensorfield_copy(TensorField *t, const TensorField *u);

extern size_t sig_tensor_size(const Sig *sig);
extern size_t sig_tensor_size_c(const Sig *sig);
extern size_t sig_slice_size(const Sig *sig);
extern size_t sig_slice_count(const Sig *sig);
extern size_t sig_tensor_count(const Sig *sig);
extern size_t sig_tensorfield_size(const Sig *sig);

extern void sig_check_max(const Sig *sig);
extern void sig_check_match(const Sig *sig1, const Sig *sig2);

extern FNumber * DEVPTR sig_get_weight(const Sig *sig);
extern FNumber *sig_get_weight_host(const Sig *sig);
extern FNumber *tensorfield_get_weight(const TensorField *t);

#endif /* _TENSOR_H */
