/*
 * dticode/src/vector.c
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */


#include <libtu/objp.h>
#include <libtu/misc.h>
#include <libtu/output.h>
#include "number.h"
#include "vector.h"

void vector_scaleplus_dummy(Vector *a, 
                            const Vector *b, 
                            const Vector *c, 
                            FNumber gamma, 
                            FNumber tau, 
                            FNumber sigma)
{
}


DYNFUN void vector_scaleplus(Vector *a, 
                             const Vector *b, 
                             const Vector *c, 
                             FNumber gamma, 
                             FNumber tau, 
                             FNumber sigma)
{
    CALL_DYN(vector_scaleplus, a, (a, b, c, gamma, tau, sigma));
}

DYNFUN void vector_scale(Vector *a, FNumber gamma)
{
    CALL_DYN(vector_scale, a, (a, gamma));
}

static void vector_scale_default(Vector *a, FNumber gamma)
{
    vector_scaleplus(a, a, a, gamma, 0, 1);
}

DYNFUN void vector_copy(Vector *a, const Vector *b)
{
    CALL_DYN(vector_copy, a, (a, b));
}

static void vector_copy_default(Vector *a, const Vector *b)
{
    vector_scaleplus(a, b, b, 1, 0, 1);
}

void vector_zero_dummy(Vector *a)
{
}

DYNFUN void vector_zero(Vector *a)
{
    CALL_DYN(vector_zero, a, (a));
}

DYNFUN FNumber vector_iprod(const Vector *a, const Vector *b)
{
    FNumber ret=0;
    
    CALL_DYN_RET(ret, FNumber, vector_iprod, a, (a, b));
    
    return ret;
}

DYNFUN FNumber vector_dist(const Vector *a, const Vector *b)
{
    FNumber ret=0;
    
    CALL_DYN_RET(ret, FNumber, vector_dist, a, (a, b));
    
    return ret;
}

DYNFUN FNumber vector_norm(const Vector *a)
{
    FNumber ret=0;
    
    CALL_DYN_RET(ret, FNumber, vector_norm, a, (a));
    
    return ret;
}

DYNFUN FNumber vector_dist22(const Vector *a, const Vector *b)
{
    FNumber ret=0;
    
    CALL_DYN_RET(ret, FNumber, vector_dist22, a, (a, b));
    
    return ret;
}

DYNFUN FNumber vector_norm22(const Vector *a)
{
    FNumber ret=0;
    
    CALL_DYN_RET(ret, FNumber, vector_norm22, a, (a));
    
    return ret;
}

FNumber vector_dist22_default(const Vector *a, const Vector *b)
{
    return (vector_iprod(a, a)+vector_iprod(b, b)-2*vector_iprod(a, b))/2;
}

FNumber vector_norm22_default(const Vector *a)
{
    return vector_iprod(a, a)/2;
}

FNumber vector_dist_default(const Vector *a, const Vector *b)
{
    return fn_sqrt(2*vector_dist22(a, b));
}

FNumber vector_norm_default(const Vector *a)
{
    return fn_sqrt(2*vector_norm22(a));
}


DYNFUN void vector_save(const Vector *a, const char *outfile)
{
    CALL_DYN(vector_save, a, (a, outfile));
}

DYNFUN bool vector_load(Vector *a, const char *outfile)
{
    bool ret=FALSE;
    
    CALL_DYN_RET(ret, bool, vector_load, a, (a, outfile));
    
    return ret;
}

bool vector_init(Vector *a)
{
    return TRUE;
}

Vector *alloc_dummyvector()
{
    CREATEOBJ_IMPL(Vector, vector, (p));
}

void vector_deinit(Vector *a)
{
}

static DynFunTab vector_dynfuntab[]={
    {vector_scale, vector_scale_default},
    {vector_copy, vector_copy_default},
    {vector_zero, vector_zero_dummy},
    {vector_scaleplus, vector_scaleplus_dummy},
    {(DynFun*)vector_dist, (DynFun*)vector_dist_default},
    {(DynFun*)vector_norm, (DynFun*)vector_norm_default},
    {(DynFun*)vector_dist22, (DynFun*)vector_dist22_default},
    {(DynFun*)vector_norm22, (DynFun*)vector_norm22_default},
    END_DYNFUNTAB,
};

IMPLCLASS(Vector, Obj, vector_deinit, vector_dynfuntab);

