/*
 * dticode/src/vector.h
 *
 * Written 2013 by Tuomo Valkonen <tuomov@iki.fi>, University of Cambridge.
 *
 */


#ifndef _VECTOR_H
#define _VECTOR_H

#include <libtu/obj.h>
#include "number.h"

INTRCLASS(Vector);

DECLCLASS(Vector){
    Obj obj;
};

DYNFUN void vector_scaleplus(Vector *a, 
                             const Vector *b, 
                             const Vector *c, 
                             FNumber gamma, 
                             FNumber tau, 
                             FNumber sigma);

DYNFUN void vector_scale(Vector *a, FNumber gamma);

DYNFUN void vector_copy(Vector *a, const Vector *b);
DYNFUN void vector_zero(Vector *a);

DYNFUN FNumber vector_iprod(const Vector *a, const Vector *b);
DYNFUN FNumber vector_dist(const Vector *a, const Vector *b);
DYNFUN FNumber vector_norm(const Vector *a);
DYNFUN FNumber vector_dist22(const Vector *a, const Vector *b); // |a-b|^2/2
DYNFUN FNumber vector_norm22(const Vector *a); // |a|^2/2

DYNFUN void vector_save(const Vector *a, const char *outfile);
DYNFUN bool vector_load(Vector *a, const char *infile);

extern void vector_deinit(Vector *a);
extern bool vector_init(Vector *a);


extern Vector *alloc_dummyvector();

#endif /* _VECTOR_H */
