%
% dticode/matlab/tensorreg_export_raw.m
%
% Written 2012 by Tuomo Valkonen <tuomov@iki.fi>, University of Graz.
%
function tensorreg_export_po(name, data, err)
    w=size(data.I, 1);
    h=size(data.I, 2);
    d=size(data.I, 3);
    m=size(data.I, 4);
    eps=20;
    infilename=[name, '.doubles'];
    amatname=[name, '-amat.doubles'];
    maskname=[name, '-mask.mat'];
    
    % Construct A=[B_1; ...; B_n], where
    % B_i=bval(i) * (bvec(i) (*) bvec(i))
    A=[];
    W=[];
    bnz=[];
    bz=[];
    bz_current=[];
    jj=[kx([0 0]) kx([0 1]) kx([1 1]) kx([0 2]) kx([1 2]) kx([2 2])];
    for i=1:m
        if data.bval(i)==0
            bz_current=i;
        else
            assert(~isempty(bz_current));
            bv=data.bvec(:, i);
            bm=data.bval(i)*bv*bv';
            bs(jj)=[bm(1, 1) 2*bm(1,2) bm(2,2) 2*bm(1,3) 2*bm(2,3) bm(3,3)];
            % Normalise for numerical accuracy.
            nn=norm(bs);
            A=[A; -bs/nn];
            W=[W; nn];
            bz=[bz, bz_current];
            bnz=[bnz, i];
        end
    end

    % Compute c=[log(D_1)/log(D_0(1)); ...; log(D_N)/log(D_0(N))]
    % so the fidelity function is ||c-Au||^2
    im=reshape(shiftdim(data.I, 3), m, w*h*d);
    
    c=log(im(bnz, :)./im(bz, :))./repmat(W, 1, w*h*d);
    % constraints computation based on confidence interval
    lb=c;
    rb=c;
     
    for i=1:length(bz)
        
        I=im(bnz(i), :);
        I0=im(bz(i), :);
        
        [delta_l,delta_r]=dti_conf_int(I,[w h d],err);
        [delta0_l,delta0_r]=dti_conf_int(I0,[w h d],err);
       
        %'here'
        %log(Il./I0r)
        %log(Ir./I0l)
        %lb(i,:)=log(Il./I0r)./W(i);
        %rb(i,:)=log(Ir./I0l)./W(i);            
        lb(i,:)=log((im(bnz(i), :)-delta_l)./(im(bz(i), :)+delta0_r))./repmat(W(i), 1, w*h*d);
        rb(i,:)=log((im(bnz(i), :)+delta_r)./(im(bz(i), :)-delta0_l))./repmat(W(i), 1, w*h*d);   
        
        
    end;
    
    %cutting out bad values
    
    bad=~isfinite(im);
    if any(any(bad))
        warning on;
        warning('NaN or Inf entries -- zeroing');
        c(bad)=0;
    end
    
    for i=1:length(bz)
        
        for j=1:w*h*d
            if not(isreal(c(i,j)))
                c(i,j)=0;       
            end;
            if not(isreal(lb(i,j)))
                lb(i,j)=-eps;       
            end;
            if not(isreal(rb(i,j)))
                rb(i,j)=eps;       
            end;
            if i==12
                fprintf('%f\n',j/(w*h*d))
            end;
        end;
     fprintf('%f\n',i/length(bz))    
    end;
    
    rawsig=[size(c, 1) 1 w h d];
    sig=[3 2 w h d];
    
    %ts=tsize_sym(sig);
    %res=zeros(ts, w*h*d);
    %ok=all(isreal(c) & isfinite(c), 1);
    %res(:, ok)=A \ c(:, ok);

    % calculate mask
    size(data.I)
    avg=reshape(sum(abs(data.I), 4), 1, w*h*d);
    threshold=0.10*max(avg);
    mask=avg>threshold;
    save(maskname, 'mask');
   
    % Write the results
    write_doubles(infilename, c);
    write_doubles([name '-lb-' int2str(err*100) '.doubles'], lb);
    write_doubles([name '-rb-' int2str(err*100) '.doubles'], rb);
    write_doubles(amatname, A);

    parf=fopen([name, '.params'], 'w');
    
    fprintf(parf, '-dim %d\n', sig(1));
    fprintf(parf, '-order %d\n', sig(2));
    fprintf(parf, '-fdim %d\n', rawsig(1));
    fprintf(parf, '-forder %d\n', rawsig(2));
    fprintf(parf, '-width %d\n', sig(3));
    fprintf(parf, '-height %d\n', sig(4));
    fprintf(parf, '-depth %d\n', sig(5));
    fprintf(parf, '-amat %s\n', amatname);
    fprintf(parf, '-infile %s\n', infilename);
    
    fclose(parf);
end

function res=kx(a)
    res=index_sym(3, a);
end
